# ----------------------------------------------------------------------------
# Copyright (c) 2016-2023, QIIME 2 development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

import os
import gc
import re
import shutil
import unittest
from unittest.mock import patch
import tempfile
import zipfile
import bibtexparser as bp

from click.testing import CliRunner
from qiime2 import Artifact, Metadata
from qiime2.core.testing.util import get_dummy_plugin
from qiime2.metadata.base import SUPPORTED_COLUMN_TYPES
from qiime2.core.cache import Cache
from qiime2.sdk.result import Result
from qiime2.sdk.plugin_manager import PluginManager

from q2cli.util import load_metadata
from q2cli.builtin.tools import tools
from q2cli.commands import RootCommand
from q2cli.core.usage import ReplayCLIUsage


class TestCastMetadata(unittest.TestCase):
    def setUp(self):
        self.runner = CliRunner()
        self.tempdir = tempfile.mkdtemp(prefix='qiime2-q2cli-test-temp-')

        self.metadata_file = os.path.join(
                self.tempdir, 'metadata.tsv')
        with open(self.metadata_file, 'w') as f:
            f.write('id\tnumbers\tstrings\n0\t42\tabc\n1\t-1.5\tdef')

        self.cast_metadata_dump = \
            ('id\tnumbers\tstrings\n#q2:types\tcategorical\tcategorical\n0\t42'
             '\tabc\n1\t-1.5\tdef\n\n')

        self.output_file = os.path.join(
                self.tempdir, 'test_output.tsv')

    def test_input_invalid_column_type(self):
        result = self.runner.invoke(
            tools, ['cast-metadata', self.metadata_file, '--cast',
                    'numbers:foo', '--output-file', self.output_file])

        self.assertNotEqual(result.exit_code, 0)
        self.assertIn('Unknown column type provided.', result.output)

    def test_input_duplicate_columns(self):
        result = self.runner.invoke(
            tools, ['cast-metadata', self.metadata_file, '--cast',
                    'numbers:numerical', '--cast', 'numbers:categorical',
                    '--output-file', self.output_file])

        self.assertNotEqual(result.exit_code, 0)
        self.assertIn(
            '"numbers" appears in cast more than once.', result.output)

    def test_input_invalid_cast_format_missing_colon(self):
        result = self.runner.invoke(
            tools, ['cast-metadata', self.metadata_file, '--cast', 'numbers',
                    '--output-file', self.output_file])

        self.assertNotEqual(result.exit_code, 0)
        self.assertIn('Missing `:` in --cast numbers', result.output)

    def test_input_invalid_cast_format_extra_colon(self):
        result = self.runner.invoke(
            tools, ['cast-metadata', self.metadata_file, '--cast', 'numbers::',
                    '--output-file', self.output_file])

        self.assertNotEqual(result.exit_code, 0)
        self.assertIn('Incorrect number of fields in --cast numbers::',
                      result.output)
        self.assertIn('Observed 3', result.output)

    def test_error_on_extra(self):
        result = self.runner.invoke(
            tools, ['cast-metadata', self.metadata_file, '--cast',
                    'extra:numeric', '--output-file', self.output_file])

        self.assertNotEqual(result.exit_code, 0)
        self.assertIn(
            "The following cast columns were not found within the"
            " metadata: extra", result.output)

    def test_error_on_missing(self):
        result = self.runner.invoke(
            tools, ['cast-metadata', self.metadata_file, '--cast',
                    'numbers:categorical', '--error-on-missing',
                    '--output-file', self.output_file])

        self.assertNotEqual(result.exit_code, 0)
        self.assertIn(
            "The following columns within the metadata"
            " were not provided in the cast: strings",
            result.output)

    def test_extra_columns_removed(self):
        result = self.runner.invoke(
            tools, ['cast-metadata', self.metadata_file, '--cast',
                    'numbers:categorical', '--cast', 'extra:numeric',
                    '--ignore-extra', '--output-file', self.output_file])

        self.assertEqual(result.exit_code, 0)
        casted_metadata = load_metadata(self.output_file)
        self.assertNotIn('extra', casted_metadata.columns.keys())

    def test_complete_successful_run(self):
        result = self.runner.invoke(
            tools, ['cast-metadata', self.metadata_file, '--cast',
                    'numbers:categorical', '--output-file', self.output_file])

        self.assertEqual(result.exit_code, 0)
        input_metadata = load_metadata(self.metadata_file)
        self.assertEqual('numeric', input_metadata.columns['numbers'].type)

        casted_metadata = load_metadata(self.output_file)
        self.assertEqual('categorical',
                         casted_metadata.columns['numbers'].type)

    def test_write_to_stdout(self):
        result = self.runner.invoke(
            tools, ['cast-metadata', self.metadata_file, '--cast',
                    'numbers:categorical'])

        self.assertEqual(result.exit_code, 0)
        self.assertEqual(self.cast_metadata_dump, result.output)

    def test_valid_column_types(self):
        result = self.runner.invoke(tools, ['cast-metadata', '--help'])
        for col_type in SUPPORTED_COLUMN_TYPES:
            self.assertIn(col_type, result.output)


class TestInspectMetadata(unittest.TestCase):
    def setUp(self):
        dummy_plugin = get_dummy_plugin()

        self.runner = CliRunner()
        self.tempdir = tempfile.mkdtemp(prefix='qiime2-q2cli-test-temp-')

        self.metadata_file_mixed_types = os.path.join(
                self.tempdir, 'metadata-mixed-types.tsv')
        with open(self.metadata_file_mixed_types, 'w') as f:
            f.write('id\tnumbers\tstrings\n0\t42\tabc\n1\t-1.5\tdef\n')

        self.bad_metadata_file = os.path.join(
                self.tempdir, 'bad-metadata.tsv')
        with open(self.bad_metadata_file, 'w') as f:
            f.write('wrong\tnumbers\tstrings\nid1\t42\tabc\nid2\t-1.5\tdef\n')

        self.metadata_artifact = os.path.join(self.tempdir, 'metadata.qza')
        Artifact.import_data(
            'Mapping', {'a': 'dog', 'b': 'cat'}).save(self.metadata_artifact)

        self.ints1 = os.path.join(self.tempdir, 'ints1.qza')
        ints1 = Artifact.import_data(
            'IntSequence1', [0, 42, 43], list)
        ints1.save(self.ints1)

        self.ints2 = os.path.join(self.tempdir, 'ints')
        ints1.export_data(self.ints2)

        self.viz = os.path.join(self.tempdir, 'viz.qzv')
        most_common_viz = dummy_plugin.actions['most_common_viz']
        self.viz = most_common_viz(ints1).visualization.save(self.viz)

    def tearDown(self):
        shutil.rmtree(self.tempdir)

    def test_artifact_w_metadata(self):
        result = self.runner.invoke(
            tools, ['inspect-metadata', self.metadata_artifact])

        self.assertEqual(result.exit_code, 0)
        self.assertIn('COLUMN NAME  TYPE', result.output)
        self.assertIn("===========  ===========", result.output)
        self.assertIn("a  categorical", result.output)
        self.assertIn("b  categorical", result.output)
        self.assertIn("IDS:  1", result.output)
        self.assertIn("COLUMNS:  2", result.output)

    def test_artifact_no_metadata(self):
        result = self.runner.invoke(tools, ['inspect-metadata', self.ints1])

        self.assertEqual(result.exit_code, 1)
        self.assertIn("IntSequence1 cannot be viewed as QIIME 2 metadata",
                      result.output)

    def test_visualization(self):
        # make a viz first:
        qiime_cli = RootCommand()
        command = qiime_cli.get_command(ctx=None, name='dummy-plugin')
        # build output parameter arguments and expected output file names
        viz_path = os.path.join(self.tempdir, 'viz.qzv')
        result = self.runner.invoke(
            command, ['most-common-viz', '--i-ints', self.ints1,
                      '--o-visualization', viz_path, '--verbose'])

        result = self.runner.invoke(tools, ['inspect-metadata', viz_path])

        self.assertEqual(result.exit_code, 1)
        self.assertIn("Visualizations cannot be viewed as QIIME 2 metadata",
                      result.output)

    def test_metadata_file(self):
        result = self.runner.invoke(
            tools, ['inspect-metadata', self.metadata_file_mixed_types])

        self.assertEqual(result.exit_code, 0)
        self.assertIn('COLUMN NAME  TYPE', result.output)
        self.assertIn("===========  ===========", result.output)
        self.assertIn("numbers  numeric", result.output)
        self.assertIn("strings  categorical", result.output)
        self.assertIn("IDS:  2", result.output)
        self.assertIn("COLUMNS:  2", result.output)

    def test_bad_metadata_file(self):
        result = self.runner.invoke(
            tools, ['inspect-metadata', self.bad_metadata_file])

        self.assertEqual(result.exit_code, 1)
        self.assertIn("'wrong'", result.output)

    def test_tsv(self):
        result = self.runner.invoke(tools, [
            'inspect-metadata', self.metadata_file_mixed_types, '--tsv'])

        self.assertEqual(result.exit_code, 0)
        self.assertIn('COLUMN NAME\tTYPE', result.output)
        self.assertIn("numbers\tnumeric", result.output)
        self.assertIn("strings\tcategorical", result.output)

        self.assertNotIn("=", result.output)
        self.assertNotIn("IDS:", result.output)
        self.assertNotIn("COLUMNS:", result.output)

    def test_merged_metadata(self):
        result = self.runner.invoke(tools, [
            'inspect-metadata',
            self.metadata_artifact,
            self.metadata_file_mixed_types])

        self.assertEqual(result.exit_code, 0)
        self.assertIn('COLUMN NAME  TYPE', result.output)
        self.assertIn("===========  ===========", result.output)
        self.assertIn("a  categorical", result.output)
        self.assertIn("b  categorical", result.output)
        self.assertIn("numbers  numeric", result.output)
        self.assertIn("strings  categorical", result.output)
        self.assertIn("IDS:  1", result.output)  # only 1 ID is shared
        self.assertIn("COLUMNS:  4", result.output)

    def test_export_to_dir_w_format(self):
        output_path = os.path.join(self.tempdir, 'output')
        result = self.runner.invoke(tools, [
            'export', '--input-path', self.ints1, '--output-path', output_path,
            '--output-format', 'IntSequenceDirectoryFormat'
        ])

        self.assertEqual(result.exit_code, 0)
        self.assertTrue(os.path.isdir(output_path))

    def test_export_to_dir_no_format(self):
        output_path = os.path.join(self.tempdir, 'output')
        self.runner.invoke(tools, [
            'export', '--input-path', self.viz, '--output-path', output_path
        ])

        self.assertTrue(os.path.isdir(output_path))
        self.assertIn('index.html', os.listdir(output_path))
        self.assertIn('index.tsv', os.listdir(output_path))

    def test_export_to_file(self):
        output_path = os.path.join(self.tempdir, 'output')
        result = self.runner.invoke(tools, [
            'export', '--input-path', self.ints1, '--output-path', output_path,
            '--output-format', 'IntSequenceFormatV2'
            ])

        with open(output_path, 'r') as f:
            file = f.read()
        self.assertEqual(result.exit_code, 0)
        self.assertIn('0', file)
        self.assertIn('42', file)
        self.assertIn('43', file)

    def test_export_to_file_creates_directories(self):
        output_path = os.path.join(self.tempdir, 'somewhere', 'output')
        result = self.runner.invoke(tools, [
            'export', '--input-path', self.ints1, '--output-path', output_path,
            '--output-format', 'IntSequenceFormatV2'
            ])

        with open(output_path, 'r') as f:
            file = f.read()
        self.assertEqual(result.exit_code, 0)
        self.assertIn('0', file)
        self.assertIn('42', file)
        self.assertIn('43', file)

    def test_export_visualization_to_dir(self):
        output_path = os.path.join(self.tempdir, 'output')
        self.runner.invoke(tools, [
            'export', '--input-path', self.viz, '--output-path', output_path
        ])

        self.assertIn('index.html', os.listdir(output_path))
        self.assertIn('index.tsv', os.listdir(output_path))
        self.assertTrue(os.path.isdir(output_path))

    def test_export_visualization_w_format(self):
        output_path = os.path.join(self.tempdir, 'output')
        result = self.runner.invoke(tools, [
            'export', '--input-path', self.viz, '--output-path', output_path,
            '--output-format', 'IntSequenceDirectoryFormat'
        ])

        self.assertEqual(result.exit_code, 1)
        self.assertIn('visualization', result.output)
        self.assertIn('--output-format', result.output)

    def test_export_path_file_is_replaced(self):
        output_path = os.path.join(self.tempdir, 'output')
        with open(output_path, 'w') as file:
            file.write('HelloWorld')
        self.runner.invoke(tools, [
            'export', '--input-path', self.ints1, '--output-path', output_path,
            '--output-format', 'IntSequenceFormatV2'
        ])
        with open(output_path, 'r') as f:
            file = f.read()
        self.assertNotIn('HelloWorld', file)

    def test_export_to_file_with_format_success_message(self):
        output_path = os.path.join(self.tempdir, 'output.int')
        result = self.runner.invoke(tools, [
            'export', '--input-path', self.ints1, '--output-path', output_path,
            '--output-format', 'IntSequenceFormatV2'
            ])
        success = 'Exported %s as IntSequenceFormatV2 to file %s\n' % (
                   self.ints1, output_path)
        self.assertEqual(success, result.output)

    def test_export_to_dir_without_format_success_message(self):
        output_path = os.path.join(self.tempdir, 'output')
        result = self.runner.invoke(tools, [
            'export', '--input-path', self.ints1, '--output-path', output_path
            ])
        success = 'Exported %s as IntSequenceDirectoryFormat to '\
                  'directory %s\n' % (self.ints1, output_path)
        self.assertEqual(success, result.output)

    def test_export_visualization_to_dir_success_message(self):
        output_path = os.path.join(self.tempdir, 'output')
        result = self.runner.invoke(tools, [
            'export', '--input-path', self.viz, '--output-path', output_path
        ])
        success = 'Exported %s as Visualization to '\
                  'directory %s\n' % (self.viz, output_path)
        self.assertEqual(success, result.output)

    def test_extract_to_dir_success_message(self):
        result = self.runner.invoke(tools, [
            'extract', '--input-path', self.ints1,
            '--output-path', self.tempdir
            ])
        success = 'Extracted %s to directory %s' % (self.ints1, self.tempdir)
        self.assertIn(success, result.output)

    def test_import_from_directory_without_format_success_message(self):
        output_path = os.path.join(self.tempdir, 'output.qza')
        result = self.runner.invoke(tools, [
            'import', '--input-path', self.ints2, '--type', 'IntSequence1',
            '--output-path', output_path
            ])
        success = 'Imported %s as IntSequenceDirectoryFormat to '\
                  '%s\n' % (self.ints2, output_path)
        self.assertEqual(success, result.output)

    def test_import_from_file_with_format_success_message(self):
        output_path = os.path.join(self.tempdir, 'output.qza')
        result = self.runner.invoke(tools, [
            'import', '--input-path', os.path.join(self.ints2, 'ints.txt'),
            '--type', 'IntSequence1',
            '--output-path', output_path,
            '--input-format', 'IntSequenceFormat'
        ])
        success = 'Imported %s as IntSequenceFormat to '\
                  '%s\n' % (os.path.join(self.ints2, 'ints.txt'), output_path)
        self.assertEqual(success, result.output)


class TestExportToFileFormat(TestInspectMetadata):
    def setUp(self):
        super().setUp()
        # Working directory is changed to temp directory to prevent cluttering
        # the repo directory with test files
        self.current_dir = os.getcwd()
        os.chdir(self.tempdir)

    def tearDown(self):
        super().tearDown()
        os.chdir(self.current_dir)

    def test_export_file_format(self):
        output_path = os.path.join(os.getcwd(), 'output')
        result = self.runner.invoke(tools, [
            'export', '--input-path', self.ints1, '--output-path', output_path,
            '--output-format', 'IntSequenceFormat'
        ])

        success = 'Exported %s as IntSequenceFormat to file %s\n' % \
                  (self.ints1, output_path)
        self.assertEqual(success, result.output)

    def test_export_dir_format(self):
        result = self.runner.invoke(tools, [
            'export', '--input-path', self.ints1, '--output-path', os.getcwd(),
            '--output-format', 'IntSequenceDirectoryFormat'
        ])

        success = 'Exported %s as IntSequenceDirectoryFormat to directory ' \
                  '%s\n' % (self.ints1, os.getcwd())
        self.assertEqual(success, result.output)

    def test_export_dir_format_nested(self):
        output_path = os.path.join(os.getcwd(), 'output')
        result = self.runner.invoke(tools, [
            'export', '--input-path', self.ints1, '--output-path', output_path,
            '--output-format', 'IntSequenceDirectoryFormat'
        ])

        success = 'Exported %s as IntSequenceDirectoryFormat to directory ' \
                  '%s\n' % (self.ints1, output_path)
        self.assertEqual(success, result.output)

    def test_export_to_filename_without_path(self):
        output_path = 'output'
        result = self.runner.invoke(tools, [
            'export', '--input-path', self.viz, '--output-path', output_path
        ])
        success = 'Exported %s as Visualization to '\
                  'directory %s\n' % (self.viz, output_path)
        self.assertEqual(success, result.output)


class TestImport(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.runner = CliRunner()

        cls.tempdir = tempfile.mkdtemp(prefix='qiime2-q2cli-test-temp-')

        cls.in_dir1 = os.path.join(cls.tempdir, 'input1')
        os.mkdir(cls.in_dir1)
        with open(os.path.join(cls.in_dir1, 'ints.txt'), 'w') as fh:
            for i in range(5):
                fh.write(f'{i}\n')
            fh.write('a\n')

        cls.in_dir2 = os.path.join(cls.tempdir, 'input2')
        os.mkdir(cls.in_dir2)
        with open(os.path.join(cls.in_dir2, 'ints.txt'), 'w') as fh:
            fh.write('1\n')
            fh.write('a\n')
            fh.write('3\n')

        cls.cache = Cache(os.path.join(cls.tempdir, 'new_cache'))

    @classmethod
    def tearDownClass(cls):
        shutil.rmtree(cls.tempdir)

    def test_import_min_validate(self):
        out_fp = os.path.join(self.tempdir, 'out1.qza')

        # import with min allows format error outside of min purview
        # (validate level min checks only first 5 items)
        result = self.runner.invoke(tools, [
            'import', '--type', 'IntSequence1', '--input-path', self.in_dir1,
            '--output-path', out_fp, '--validate-level', 'min'
        ])
        self.assertEqual(result.exit_code, 0)

        # import with max should catch all format errors, max is default
        result = self.runner.invoke(tools, [
            'import', '--type', 'IntSequence1', '--input-path',
            self.in_dir1, '--output-path', out_fp
        ])
        self.assertEqual(result.exit_code, 1)
        self.assertIn('Line 6 is not an integer', result.output)

        out_fp = os.path.join(self.tempdir, 'out2.qza')

        # import with min catches format errors within its purview
        result = self.runner.invoke(tools, [
            'import', '--type', 'IntSequence1', '--input-path',
            self.in_dir2, '--output-path', out_fp, '--validate-level', 'min'
        ])
        self.assertEqual(result.exit_code, 1)
        self.assertIn('Line 2 is not an integer', result.output)

    def test_cache_import_min_validate(self):
        # import with min allows format error outside of min purview
        # (validate level min checks only first 5 items)
        result = self.runner.invoke(tools, [
            'cache-import', '--type', 'IntSequence1', '--input-path',
            self.in_dir1, '--cache', str(self.cache.path), '--key', 'foo',
            '--validate-level', 'min'
        ])
        self.assertEqual(result.exit_code, 0)

        # import with max should catch all format errors, max is default
        result = self.runner.invoke(tools, [
            'cache-import', '--type', 'IntSequence1', '--input-path',
            self.in_dir1, '--cache', str(self.cache.path), '--key', 'foo'
        ])
        self.assertEqual(result.exit_code, 1)
        self.assertIn('Line 6 is not an integer', result.output)

        # import with min catches format errors within its purview
        result = self.runner.invoke(tools, [
            'cache-import', '--type', 'IntSequence1', '--input-path',
            self.in_dir2, '--cache', str(self.cache.path), '--key', 'foo',
            '--validate-level', 'min'
        ])
        self.assertEqual(result.exit_code, 1)
        self.assertIn('Line 2 is not an integer', result.output)


class TestCacheTools(unittest.TestCase):
    def setUp(self):
        get_dummy_plugin()

        self.runner = CliRunner()
        self.plugin_command = RootCommand().get_command(
            ctx=None, name='dummy-plugin')
        self.tempdir = \
            tempfile.TemporaryDirectory(prefix='qiime2-q2cli-test-temp-')

        self.art1 = Artifact.import_data('IntSequence1', [0, 1, 2])
        self.art2 = Artifact.import_data('IntSequence1', [3, 4, 5])
        self.art3 = Artifact.import_data('IntSequence1', [6, 7, 8])
        self.art4 = Artifact.import_data('IntSequence2', [9, 10, 11])
        self.to_import = os.path.join(self.tempdir.name, 'to_import')
        self.art1.export_data(self.to_import)
        self.cache = Cache(os.path.join(self.tempdir.name, 'new_cache'))

    def tearDown(self):
        self.tempdir.cleanup()

    def test_cache_create(self):
        cache_path = os.path.join(self.tempdir.name, 'created_cache')

        result = self.runner.invoke(
            tools, ['cache-create', '--cache', cache_path])

        success = "Created cache at '%s'\n" % cache_path
        self.assertEqual(success, result.output)
        self.assertTrue(Cache.is_cache(cache_path))

    def test_cache_remove(self):
        self.cache.save(self.art1, 'key')
        self.assertTrue('key' in self.cache.get_keys())

        result = self.runner.invoke(
            tools,
            ['cache-remove', '--cache', str(self.cache.path), '--key', 'key'])

        success = "Removed key 'key' from cache '%s'\n" % self.cache.path
        self.assertEqual(success, result.output)
        self.assertFalse('key' in self.cache.get_keys())

    def test_cache_garbage_collection(self):
        # Data referenced directly by key
        self.cache.save(self.art1, 'foo')
        # Data referenced by pool that is referenced by key
        pool = self.cache.create_pool(key='bar')
        pool.save(self.art2)
        # We will be manually deleting the keys that back these two
        self.cache.save(self.art3, 'baz')
        pool = self.cache.create_pool(key='qux')
        pool.save(self.art4)

        # What we expect to see before and after gc
        expected_pre_gc_contents = \
            set(('./VERSION', 'keys/foo', 'keys/bar',
                 'keys/baz', 'keys/qux',
                 f'pools/bar/{self.art2.uuid}',
                 f'pools/qux/{self.art4.uuid}',
                 f'data/{self.art1.uuid}', f'data/{self.art2.uuid}',
                 f'data/{self.art3.uuid}', f'data/{self.art4.uuid}'))

        expected_post_gc_contents = \
            set(('./VERSION', 'keys/foo', 'keys/bar',
                 f'pools/bar/{self.art2.uuid}',
                 f'data/{self.art1.uuid}', f'data/{self.art2.uuid}'))

        # Assert cache looks how we want pre gc
        pre_gc_contents = _get_cache_contents(self.cache)
        self.assertEqual(expected_pre_gc_contents, pre_gc_contents)

        # Delete keys
        self.cache.remove(self.cache.keys / 'baz')
        self.cache.remove(self.cache.keys / 'qux')

        # Make sure Python's garbage collector gets the process pool symlinks
        # to the artifact that was keyed on baz and the one in the qux pool
        gc.collect()
        result = self.runner.invoke(
            tools,
            ['cache-garbage-collection', '--cache', str(self.cache.path)])

        success = "Ran garbage collection on cache at '%s'\n" % self.cache.path
        self.assertEqual(success, result.output)

        # Assert cache looks how we want post gc
        post_gc_contents = _get_cache_contents(self.cache)
        self.assertEqual(expected_post_gc_contents, post_gc_contents)

    def test_cache_store(self):
        artifact = os.path.join(self.tempdir.name, 'artifact.qza')
        self.art1.save(artifact)

        result = self.runner.invoke(
            tools, ['cache-store', '--cache', str(self.cache.path),
                    '--artifact-path', artifact, '--key', 'key'])

        success = "Saved the artifact '%s' to the cache '%s' under the key " \
            "'key'\n" % (artifact, self.cache.path)
        self.assertEqual(success, result.output)

    def test_cache_fetch(self):
        artifact = os.path.join(self.tempdir.name, 'artifact.qza')
        self.cache.save(self.art1, 'key')

        result = self.runner.invoke(
            tools, ['cache-fetch', '--cache', str(self.cache.path),
                    '--key', 'key', '--output-path', artifact])

        success = "Loaded artifact with the key 'key' from the cache '%s' " \
            "and saved it to the file '%s'\n" % (self.cache.path, artifact)
        self.assertEqual(success, result.output)

    def test_cache_roundtrip(self):
        in_artifact = os.path.join(self.tempdir.name, 'in_artifact.qza')
        out_artifact = os.path.join(self.tempdir.name, 'out_artifact.qza')

        self.art1.save(in_artifact)

        result = self.runner.invoke(
            tools, ['cache-store', '--cache', str(self.cache.path),
                    '--artifact-path', in_artifact, '--key', 'key'])

        success = "Saved the artifact '%s' to the cache '%s' under the key " \
            "'key'\n" % (in_artifact, self.cache.path)
        self.assertEqual(success, result.output)

        result = self.runner.invoke(
            tools, ['cache-fetch', '--cache', str(self.cache.path),
                    '--key', 'key', '--output-path', out_artifact])

        success = "Loaded artifact with the key 'key' from the cache '%s' " \
            "and saved it to the file '%s'\n" % (self.cache.path, out_artifact)
        self.assertEqual(success, result.output)

        artifact = Artifact.load(out_artifact)
        self.assertEqual([0, 1, 2], artifact.view(list))

    def test_cache_status(self):
        success_template = \
            "Status of the cache at the path '%s':\n\n%s\n\n%s\n"

        # Empty cache
        result = self.runner.invoke(
            tools, ['cache-status', '--cache', str(self.cache.path)])
        success = \
            success_template % (str(self.cache.path), 'No data keys in cache',
                                'No pool keys in cache')
        self.assertEqual(success, result.output)

        # Cache with only data
        in_artifact = os.path.join(self.tempdir.name, 'in_artifact.qza')
        self.art1.save(in_artifact)
        self.runner.invoke(
            tools, ['cache-store', '--cache', str(self.cache.path),
                    '--artifact-path', in_artifact, '--key', 'key'])

        result = self.runner.invoke(
            tools, ['cache-status', '--cache', str(self.cache.path)])
        data_output = 'Data keys in cache:\ndata: key -> %s' % \
            str(Result.peek(self.cache.data / str(self.art1.uuid)))
        success = \
            success_template % (str(self.cache.path), data_output,
                                'No pool keys in cache')
        self.assertEqual(success, result.output)

        # Cache with data and pool
        pool = self.cache.create_pool(key='pool')
        pool.save(self.art2)

        result = self.runner.invoke(
            tools, ['cache-status', '--cache', str(self.cache.path)])
        pool_output = 'Pool keys in cache:\npool: pool -> size = 1'
        success = \
            success_template % (str(self.cache.path), data_output,
                                pool_output)
        self.assertEqual(success, result.output)

    def test_cache_import(self):
        self.max_diff = None
        result = self.runner.invoke(
            tools, ['cache-import', '--type', 'IntSequence1', '--input-path',
                    self.to_import, '--cache', f'{self.cache.path}', '--key',
                    'foo'])
        success = 'Imported %s as IntSequenceDirectoryFormat to %s:foo\n' % \
            (self.to_import, self.cache.path)
        self.assertEqual(success, result.output)


def _get_cache_contents(cache):
    """Gets contents of cache not including contents of the artifacts
    themselves relative to the root of the cache
    """
    cache_contents = set()

    rel_keys = os.path.relpath(cache.keys, cache.path)
    rel_data = os.path.relpath(cache.data, cache.path)
    rel_pools = os.path.relpath(cache.pools, cache.path)
    rel_cache = os.path.relpath(cache.path, cache.path)

    for key in os.listdir(cache.keys):
        cache_contents.add(os.path.join(rel_keys, key))

    for art in os.listdir(cache.data):
        cache_contents.add(os.path.join(rel_data, art))

    for pool in os.listdir(cache.pools):
        for link in os.listdir(os.path.join(cache.pools, pool)):
            cache_contents.add(os.path.join(rel_pools, pool, link))

    for elem in os.listdir(cache.path):
        if os.path.isfile(os.path.join(cache.path, elem)):
            cache_contents.add(os.path.join(rel_cache, elem))

    return cache_contents


class TestPeek(unittest.TestCase):
    def setUp(self):
        self.runner = CliRunner()
        self.tempdir = tempfile.mkdtemp(prefix='qiime2-q2cli-test-temp-')

        # create artifact
        self.artifact = os.path.join(self.tempdir, 'artifact.qza')
        Artifact.import_data(
            'Mapping', {'foo': 'bar'}).save(self.artifact)

        # create visualization
        qiime_cli = RootCommand()
        command = qiime_cli.get_command(ctx=None, name='dummy-plugin')
        self.viz = os.path.join(self.tempdir, 'viz.qzv')

        self.ints = os.path.join(self.tempdir, 'ints.qza')
        ints = Artifact.import_data(
            'IntSequence1', [0, 42, 43], list)
        ints.save(self.ints)

        self.runner.invoke(
            command, ['most-common-viz', '--i-ints', self.ints,
                      '--o-visualization', self.viz, '--verbose'])

    def tearDown(self):
        shutil.rmtree(self.tempdir)

    def test_single_artifact(self):
        result = self.runner.invoke(tools, ['peek', self.artifact])
        self.assertEqual(result.exit_code, 0)
        self.assertIn("UUID:", result.output)
        self.assertIn("Type:", result.output)
        self.assertIn("Data format:", result.output)
        self.assertEqual(result.output.count('\n'), 3)

    def test_single_visualization(self):
        result = self.runner.invoke(tools, ['peek', self.viz])
        self.assertEqual(result.exit_code, 0)
        self.assertIn("UUID:", result.output)
        self.assertIn("Type:", result.output)
        self.assertNotIn("Data format:", result.output)
        self.assertEqual(result.output.count('\n'), 2)

    def test_artifact_and_visualization(self):
        result = self.runner.invoke(tools, ['peek', self.artifact, self.viz])
        self.assertEqual(result.exit_code, 0)
        self.assertIn("UUID", result.output)
        self.assertIn("Type", result.output)
        self.assertIn("Data Format", result.output)
        self.assertIn("N/A", result.output)
        self.assertEqual(result.output.count('\n'), 3)

    def test_single_file_tsv(self):
        result = self.runner.invoke(tools, ['peek', '--tsv', self.artifact])
        self.assertIn("Filename\tType\tUUID\tData Format\n", result.output)
        self.assertIn("artifact.qza", result.output)
        self.assertEqual(result.output.count('\t'), 6)
        self.assertEqual(result.output.count('\n'), 2)

    def test_multiple_file_tsv(self):
        result = self.runner.invoke(tools, ['peek', '--tsv', self.artifact,
                                            self.viz])
        self.assertIn("Filename\tType\tUUID\tData Format\n", result.output)
        self.assertIn("artifact.qza", result.output)
        self.assertIn("viz.qzv", result.output)
        self.assertEqual(result.output.count('\t'), 9)
        self.assertEqual(result.output.count('\n'), 3)


class TestListTypes(unittest.TestCase):
    def setUp(self):
        self.runner = CliRunner()
        self.pm = PluginManager()

    def tearDown(self):
        pass

    def test_list_all_types(self):
        result = self.runner.invoke(tools, ['list-types'])
        self.assertEqual(result.exit_code, 0)

        for name, artifact_class_record in self.pm.artifact_classes.items():
            self.assertIn(name, result.output)
            self.assertIn(artifact_class_record.description, result.output)

    def test_list_types_fuzzy(self):
        types = list(self.pm.artifact_classes)[:5]
        result = self.runner.invoke(tools, ['list-types', *types])
        self.assertEqual(result.exit_code, 0)

        # split on \n\n because types and their description are separated
        # by two newlines
        # len - 1 because split includes '' for the last \n\n split
        self.assertGreaterEqual(len(result.output.split('\n\n')) - 1,
                                len(types))

    def test_list_types_strict(self):
        types = list(self.pm.artifact_classes)[:5]
        result = self.runner.invoke(tools, ['list-types', '--strict', *types])
        self.assertEqual(result.exit_code, 0)
        self.assertEqual(len(result.output.split('\n\n')) - 1, len(types))

        result = self.runner.invoke(tools, ['list-types', '--strict',
                                            types[0] + 'x'])
        self.assertEqual(result.exit_code, 0)
        self.assertEqual(len(result.output), 0)

        result = self.runner.invoke(tools, ['list-types', '--strict', *types,
                                            types[0] + 'x'])
        self.assertEqual(result.exit_code, 0)
        self.assertEqual(len(result.output.split('\n\n')) - 1, len(types))

    def test_list_types_tsv(self):
        result = self.runner.invoke(tools, ['list-types', '--tsv'])
        self.assertEqual(result.exit_code, 0)

        # len - 1 because \n split produces a final ''
        self.assertEqual(len(result.output.split('\n')) - 1,
                         len(self.pm.artifact_classes))

        no_description_count = 0
        for name, artifact_class_record in self.pm.artifact_classes.items():
            self.assertIn(name, result.output)
            self.assertIn(artifact_class_record.description, result.output)
            if artifact_class_record.description == '':
                no_description_count += 1

        self.assertEqual(no_description_count, result.output.count('\t\n'))


class TestListFormats(unittest.TestCase):
    def setUp(self):
        self.runner = CliRunner()
        self.pm = PluginManager()

    def tearDown(self):
        pass

    def test_list_all_importable_formats(self):
        result = self.runner.invoke(tools, ['list-formats', '--importable'])
        self.assertEqual(result.exit_code, 0)

        for name, format_record in self.pm.importable_formats.items():
            self.assertIn(name, result.output)
            docstring = format_record.format.__doc__
            if docstring:
                description = docstring.split('\n\n')[0].strip()
                for word in description:
                    self.assertIn(word.strip(), result.output)

    def test_list_all_exportable_formats(self):
        result = self.runner.invoke(tools, ['list-formats', '--exportable'])
        self.assertEqual(result.exit_code, 0)

        for name, format_record in self.pm.exportable_formats.items():
            self.assertIn(name, result.output)
            docstring = format_record.format.__doc__
            if docstring:
                description = docstring.split('\n\n')[0].strip()
                for word in description:
                    self.assertIn(word.strip(), result.output)

    def test_list_formats_fuzzy(self):
        formats = list(self.pm.importable_formats)[:5]
        result = self.runner.invoke(tools, ['list-formats', '--importable',
                                            *formats])
        self.assertEqual(result.exit_code, 0)

        # see TestListTypes.test_list_types_fuzzy
        self.assertGreaterEqual(len(result.output.split('\n\n')) - 1,
                                len(formats))

    def test_list_formats_strict(self):
        formats = list(self.pm.exportable_formats)[:5]
        result = self.runner.invoke(tools, ['list-formats', '--exportable',
                                            '--strict', *formats])
        self.assertEqual(result.exit_code, 0)
        self.assertEqual(len(result.output.split('\n\n')) - 1, len(formats))

        result = self.runner.invoke(tools, ['list-formats', '--exportable',
                                            '--strict', formats[0] + 'x'])
        self.assertEqual(result.exit_code, 0)
        self.assertEqual(len(result.output), 0)

        result = self.runner.invoke(tools, ['list-formats', '--exportable',
                                            '--strict', *formats,
                                            formats[0] + 'x'])
        self.assertEqual(result.exit_code, 0)
        self.assertEqual(len(result.output.split('\n\n')) - 1, len(formats))

    def test_list_formats_tsv(self):
        result = self.runner.invoke(tools, ['list-formats', '--importable',
                                            '--tsv'])
        self.assertEqual(result.exit_code, 0)

        # len - 1 because \n split produces a final ''
        self.assertEqual(len(result.output.split('\n')) - 1,
                         len(self.pm.importable_formats))

        no_description_count = 0
        for name, format_record in self.pm.importable_formats.items():
            self.assertIn(name, result.output)
            docstring = format_record.format.__doc__
            if docstring:
                description = docstring.split('\n\n')[0].strip()
                for word in description:
                    self.assertIn(word.strip(), result.output)

            if format_record.format.__doc__ is None:
                no_description_count += 1
        self.assertEqual(no_description_count, result.output.count('\t\n'))


class TestReplay(unittest.TestCase):
    def setUp(self):
        self.runner = CliRunner()
        self.pm = PluginManager()
        self.dp = self.pm.plugins['dummy-plugin']
        self.tempdir = tempfile.mkdtemp(prefix='q2cli-test-replay-temp-')

        # contrive artifacts with different sorts of provenance
        int_seq1 = Artifact.import_data('IntSequence1', [1, 2, 3])
        int_seq2 = Artifact.import_data('IntSequence1', [4, 5, 6])
        int_seq3 = Artifact.import_data('IntSequence2', [7, 8])
        concat_ints = self.dp.actions['concatenate_ints']
        concated_ints, = concat_ints(int_seq1, int_seq2, int_seq3, 9, 0)
        concated_ints.save(os.path.join(self.tempdir, 'concated_ints.qza'))

        outer_dir = os.path.join(self.tempdir, 'outer_dir')
        inner_dir = os.path.join(self.tempdir, 'outer_dir', 'inner_dir')
        os.mkdir(outer_dir)
        os.mkdir(inner_dir)
        shutil.copy(os.path.join(self.tempdir, 'concated_ints.qza'), outer_dir)
        int_seq = Artifact.import_data('IntSequence1', [1, 2, 3, 4])
        left_ints, _ = self.dp.actions['split_ints'](int_seq)
        left_ints.save(os.path.join(inner_dir, 'left_ints.qza'))

        mapping = Artifact.import_data('Mapping', {'qiime': 2, 'triangle': 3})
        int_seq_with_md, = self.dp.actions['identity_with_metadata'](
            int_seq1,
            mapping.view(Metadata))
        int_seq_with_md.save(os.path.join(self.tempdir, 'int_seq_with_md.qza'))

    def tearDown(self):
        shutil.rmtree(self.tempdir)

    def test_replay_provenance(self):
        in_fp = os.path.join(self.tempdir, 'concated_ints.qza')
        out_fp = os.path.join(self.tempdir, 'rendered.txt')
        result = self.runner.invoke(
            tools,
            ['replay-provenance', '--in-fp', in_fp, '--out-fp', out_fp]
        )
        self.assertEqual(result.exit_code, 0)

        with open(out_fp, 'r') as fh:
            rendered = fh.read()

        self.assertIn('qiime tools import', rendered)
        self.assertIn('--type \'IntSequence1\'', rendered)
        self.assertIn('--type \'IntSequence2\'', rendered)
        self.assertIn('--input-path <your data here>', rendered)
        self.assertIn('--output-path int-sequence1-0.qza', rendered)
        self.assertIn('--output-path int-sequence1-1.qza', rendered)
        self.assertIn('--output-path int-sequence2-0.qza', rendered)

        self.assertIn('qiime dummy-plugin concatenate-ints', rendered)
        self.assertRegex(rendered, '--i-ints[12] int-sequence1-0.qza')
        self.assertRegex(rendered, '--i-ints[12] int-sequence1-1.qza')
        self.assertIn('--i-ints3 int-sequence2-0.qza', rendered)
        self.assertIn('--p-int1 9', rendered)
        self.assertIn('--p-int2 0', rendered)
        self.assertIn('--o-concatenated-ints concatenated-ints-0.qza',
                      rendered)

    def test_replay_provenance_python(self):
        in_fp = os.path.join(self.tempdir, 'concated_ints.qza')
        out_fp = os.path.join(self.tempdir, 'rendered.txt')
        result = self.runner.invoke(
            tools,
            ['replay-provenance', '--in-fp', in_fp, '--out-fp', out_fp,
                '--usage-driver', 'python3']
        )
        self.assertEqual(result.exit_code, 0)

        with open(out_fp, 'r') as fh:
            rendered = fh.read()

        self.assertIn('from qiime2 import Artifact', rendered)
        self.assertIn('Artifact.import_data', rendered)
        self.assertIn('dummy_plugin_actions.concatenate_ints', rendered)

    def test_replay_provenance_recurse(self):
        """
        If the directory is parsed recursively, both the concated_ints.qza and
        left_ints.qza will be captured.
        """
        in_fp = os.path.join(self.tempdir, 'outer_dir')
        out_fp = os.path.join(self.tempdir, 'rendered.txt')
        result = self.runner.invoke(
            tools,
            ['replay-provenance', '--in-fp', in_fp, '--out-fp', out_fp,
                '--usage-driver', 'python3', '--recurse']
        )
        self.assertEqual(result.exit_code, 0)

        with open(out_fp, 'r') as fh:
            rendered = fh.read()

        self.assertIn('dummy_plugin_actions.concatenate_ints', rendered)
        self.assertIn('dummy_plugin_actions.split_ints', rendered)

    def test_replay_provenance_use_md_without_parse(self):
        in_fp = os.path.join(self.tempdir, 'outer_dir')
        out_fp = os.path.join(self.tempdir, 'rendered.txt')
        result = self.runner.invoke(
            tools,
            ['replay-provenance', '--in-fp', in_fp, '--out-fp', out_fp,
             '--no-parse-metadata', '--use-recorded-metadata']
        )
        self.assertEqual(result.exit_code, 1)
        self.assertIsInstance(result.exception, ValueError)
        self.assertRegex(str(result.exception),
                         'Metadata not parsed for replay')

    @patch('qiime2.sdk.util.get_available_usage_drivers',
           return_value={'cli': ReplayCLIUsage})
    def test_replay_provenance_usage_driver_not_available(self, patch):
        in_fp = os.path.join(self.tempdir, 'concated_ints.qza')
        out_fp = os.path.join(self.tempdir, 'rendered.txt')
        result = self.runner.invoke(
            tools,
            ['replay-provenance', '--in-fp', in_fp, '--out-fp', out_fp,
                '--usage-driver', 'python3']
        )
        self.assertEqual(result.exit_code, 1)
        self.assertIsInstance(result.exception, ValueError)
        self.assertIn(
            'python3 usage driver is not available', str(result.exception)
        )

    def test_replay_citations(self):
        in_fp = os.path.join(self.tempdir, 'concated_ints.qza')
        out_fp = os.path.join(self.tempdir, 'citations.bib')
        result = self.runner.invoke(
            tools,
            ['replay-citations', '--in-fp', in_fp, '--out-fp', out_fp]
        )
        self.assertEqual(result.exit_code, 0)

        with open(out_fp) as fh:
            bib_database = bp.load(fh)

        # use .*? to non-greedily match version strings
        exp = [
            r'action\|dummy-plugin:.*?\|method:concatenate_ints\|0',
            r'framework\|qiime2:.*?\|0',
            r'plugin\|dummy-plugin:.*?\|0',
            r'plugin\|dummy-plugin:.*?\|1',
            r'transformer\|dummy-plugin:.*?\|builtins:list->'
            r'IntSequenceDirectoryFormat\|0',
            r'transformer\|dummy-plugin:.*?\|builtins:list->'
            r'IntSequenceV2DirectoryFormat\|4',
            r'transformer\|dummy-plugin:.*?\|builtins:list->'
            r'IntSequenceV2DirectoryFormat\|5',
            r'transformer\|dummy-plugin:.*?\|builtins:list->'
            r'IntSequenceV2DirectoryFormat|6',
            r'transformer\|dummy-plugin:.*?\|builtins:list->'
            r'IntSequenceV2DirectoryFormat\|8',
            r'view\|dummy-plugin:.*?\|IntSequenceDirectoryFormat\|0'
        ]

        self.assertEqual(len(exp), len(bib_database.entries))

        all_records_str = ''
        for record in bib_database.entries_dict.keys():
            all_records_str += f' {record}'
        for record in exp:
            self.assertRegex(all_records_str, record)

    def test_replay_citations_no_deduplicate(self):
        in_fp = os.path.join(self.tempdir, 'concated_ints.qza')
        out_fp = os.path.join(self.tempdir, 'citations.bib')
        result = self.runner.invoke(
            tools,
            ['replay-citations', '--in-fp', in_fp, '--out-fp', out_fp,
             '--no-deduplicate']
        )
        self.assertEqual(result.exit_code, 0)

        with open(out_fp) as fh:
            bib_database = bp.load(fh)
        self.assertEqual(28, len(bib_database.entries))

        with open(out_fp) as fh:
            file_contents = fh.read()
        framework_citations = \
            re.compile(r'framework\|qiime2:.*?\|0.*' * 4, re.DOTALL)
        self.assertRegex(file_contents, framework_citations)

    def test_replay_supplement(self):
        in_fp = os.path.join(self.tempdir, 'concated_ints.qza')
        out_fp = os.path.join(self.tempdir, 'supplement.zip')
        result = self.runner.invoke(
            tools,
            ['replay-supplement', '--in-fp', in_fp, '--out-fp', out_fp]
        )
        self.assertEqual(result.exit_code, 0)
        self.assertTrue(zipfile.is_zipfile(out_fp))

        exp = {
            'supplement/',
            'supplement/python3_replay.py',
            'supplement/cli_replay.sh',
            'supplement/citations.bib'
        }
        with zipfile.ZipFile(out_fp, 'r') as zfh:
            self.assertEqual(exp, set(zfh.namelist()))

    def test_replay_supplement_with_metadata(self):
        in_fp = os.path.join(self.tempdir, 'int_seq_with_md.qza')
        out_fp = os.path.join(self.tempdir, 'supplement.zip')
        result = self.runner.invoke(
            tools,
            ['replay-supplement', '--in-fp', in_fp, '--out-fp', out_fp]
        )
        self.assertEqual(result.exit_code, 0)
        self.assertTrue(zipfile.is_zipfile(out_fp))

        exp = {
            'supplement/',
            'supplement/python3_replay.py',
            'supplement/cli_replay.sh',
            'supplement/citations.bib',
            'supplement/recorded_metadata/',
            'supplement/recorded_metadata/'
            'dummy_plugin_identity_with_metadata_0/',
            'supplement/recorded_metadata/'
            'dummy_plugin_identity_with_metadata_0/metadata_0.tsv',
        }
        with zipfile.ZipFile(out_fp, 'r') as zfh:
            self.assertEqual(exp, set(zfh.namelist()))

    def test_replay_supplement_no_metadata_dump(self):
        in_fp = os.path.join(self.tempdir, 'int_seq_with_md.qza')
        out_fp = os.path.join(self.tempdir, 'supplement.zip')
        result = self.runner.invoke(
            tools,
            ['replay-supplement', '--in-fp', in_fp, '--out-fp', out_fp,
             '--no-dump-recorded-metadata']
        )
        self.assertEqual(result.exit_code, 0)
        self.assertTrue(zipfile.is_zipfile(out_fp))

        not_exp = 'recorded_metadata/'
        with zipfile.ZipFile(out_fp, 'r') as zfh:
            self.assertNotIn(not_exp, set(zfh.namelist()))

    @patch('qiime2.sdk.util.get_available_usage_drivers', return_value={})
    def test_replay_supplement_usage_driver_not_available(self, patch):
        in_fp = os.path.join(self.tempdir, 'concated_ints.qza')
        out_fp = os.path.join(self.tempdir, 'rendered.txt')
        result = self.runner.invoke(
            tools,
            ['replay-supplement', '--in-fp', in_fp, '--out-fp', out_fp]
        )
        self.assertEqual(result.exit_code, 1)
        self.assertIsInstance(result.exception, ValueError)
        self.assertIn(
            'no available usage drivers', str(result.exception)
        )

    def test_replay_supplement_zipfile(self):
        with tempfile.TemporaryDirectory() as tempdir:
            in_fp = os.path.join(self.tempdir, 'concated_ints.qza')
            out_fp = os.path.join(tempdir, 'supplement.zip')

            result = self.runner.invoke(
                tools,
                ['replay-supplement', '--in-fp', in_fp, '--out-fp', out_fp]
            )
            self.assertEqual(result.exit_code, 0)
            self.assertTrue(zipfile.is_zipfile(out_fp))

            unzipped_path = os.path.join(tempdir, 'extracted')
            os.makedirs(unzipped_path)
            with zipfile.ZipFile(out_fp, 'r') as zfh:
                zfh.extractall(unzipped_path)

            self.assertEqual(os.listdir(unzipped_path), ['supplement'])


if __name__ == "__main__":
    unittest.main()



# ----------------------------------------------------------------------------
# Copyright (c) 2016-2023, QIIME 2 development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

import collections
import os
import pkg_resources
import re
import shlex
import textwrap
from typing import Any, Callable, Dict, List, Tuple

from qiime2 import ResultCollection
import qiime2.sdk.usage as usage
from qiime2.sdk.usage import (
    UsageVariable, Usage, UsageInputs, UsageOutputs, UsageOutputNames
)
from qiime2.sdk import Action
from qiime2.core.archive.provenance_lib.usage_drivers import (
    build_header, build_footer
)
from qiime2.core.archive.provenance_lib import ProvDAG

import q2cli.util as util
from q2cli.core.state import get_action_state
import q2cli.click.option


def write_example_data(action, output_dir):
    for example_name, example in action.examples.items():
        cli_name = util.to_cli_name(example_name)
        example_path = os.path.join(output_dir, cli_name)

        use = CLIUsage()
        example(use)

        for fn, val in use.get_example_data():
            os.makedirs(example_path, exist_ok=True)
            path = os.path.join(example_path, fn)
            val.save(path)
            try:
                hint = repr(val.type)
            except AttributeError:
                hint = 'Metadata'

            yield hint, path


def write_plugin_example_data(plugin, output_dir):
    for name, action in plugin.actions.items():
        path = os.path.join(output_dir, util.to_cli_name(name))

        yield from write_example_data(action, path)


class CLIUsageVariable(usage.UsageVariable):
    EXT = {
        'artifact': '.qza',
        'artifact_collection': '/',
        'visualization_collection': '/',
        'visualization': '.qzv',
        'metadata': '.tsv',
        'column': '',
        'format': '',
    }

    ELEMENT_EXT = {
        'artifact_collection': EXT['artifact'],
        'visualization_collection': EXT['visualization']
    }

    @property
    def ext(self):
        return self.EXT[self.var_type]

    @staticmethod
    def to_cli_name(val):
        return util.to_cli_name(val)

    def _key_helper(self, input_path, key):
        if self.var_type not in self.COLLECTION_VAR_TYPES:
            raise KeyboardInterrupt(
                f'Cannot key non-collection type {self.var_type}')

        return "%s%s%s" % (input_path, key, self.ELEMENT_EXT[self.var_type])

    def to_interface_name(self):
        if hasattr(self, '_q2cli_ref'):
            return self._q2cli_ref

        cli_name = '%s%s' % (self.name, self.ext)

        # don't disturb file names, this will break importing where QIIME 2
        # relies on specific filenames being present in a dir
        if self.var_type not in ('format', 'column'):
            cli_name = self.to_cli_name(cli_name)

        return cli_name

    def assert_has_line_matching(self, path, expression, key=None):
        if not self.use.enable_assertions:
            return

        INDENT = self.use.INDENT
        input_path = self.to_interface_name()
        expr = shlex.quote(expression)

        if key:
            input_path = self._key_helper(input_path, key)

        lines = [
            'qiime dev assert-result-data %s \\' % (input_path,),
            INDENT + '--zip-data-path %s \\' % (path,),
            INDENT + '--expression %s' % (expr,),
        ]

        self.use.recorder.extend(lines)

    def assert_output_type(self, semantic_type, key=None):
        if not self.use.enable_assertions:
            return

        INDENT = self.use.INDENT
        input_path = self.to_interface_name()

        if key:
            input_path = self._key_helper(input_path, key)

        lines = [
            'qiime dev assert-result-type %s \\' % (input_path,),
            INDENT + '--qiime-type %s' % (str(semantic_type),),
        ]

        self.use.recorder.extend(lines)


class CLIUsage(usage.Usage):
    INDENT = ' ' * 2

    def __init__(self, enable_assertions=False, action_collection_size=None):
        super().__init__()
        self.recorder = []
        self.init_data = []
        self.enable_assertions = enable_assertions
        self.action_collection_size = action_collection_size
        self.output_dir_counter = collections.defaultdict(int)

    def usage_variable(self, name, factory, var_type):
        return CLIUsageVariable(name, factory, var_type, self)

    def render(self, flush=False):
        rendered = '\n'.join(self.recorder)
        if flush:
            self.recorder = []
            self.init_data = []
        return rendered

    def init_artifact(self, name, factory):
        variable = super().init_artifact(name, factory)

        self.init_data.append(variable)

        return variable

    def init_artifact_collection(self, name, factory):
        variable = super().init_artifact_collection(name, factory)

        self.init_data.append(variable)

        return variable

    def construct_artifact_collection(self, name, members):
        variable = super().construct_artifact_collection(
            name, members
        )

        rc_dir = variable.to_interface_name()

        keys = members.keys()
        names = [name.to_interface_name() for name in members.values()]

        keys_arg = '( '
        for key in keys:
            keys_arg += f'{key} '
        keys_arg += ')'
        names_arg = '( '
        for name in names:
            names_arg += f'{name} '
        names_arg += ')'

        lines = [
            '## constructing result collection ##',
            f'rc_name={rc_dir}',
            'ext=.qza',
            f'keys={keys_arg}',
            f'names={names_arg}',
            'construct_result_collection',
            '##',
        ]
        self.recorder.extend(lines)

        return variable

    def get_artifact_collection_member(self, name, variable, key):
        accessed_variable = super().get_artifact_collection_member(
            name, variable, key
        )

        rc_dir = variable.to_interface_name()
        member_fp = os.path.join(rc_dir, f'{key}.qza')

        lines = [
            '## accessing result collection member ##',
            f'ln -s {member_fp} {accessed_variable.to_interface_name()}',
            '##',
        ]
        self.recorder.extend(lines)

        return variable

    def import_from_format(self, name, semantic_type,
                           variable, view_type=None):
        imported_var = super().import_from_format(
            name, semantic_type, variable, view_type=view_type)

        in_fp = variable.to_interface_name()
        out_fp = imported_var.to_interface_name()

        lines = [
            'qiime tools import \\',
            self.INDENT + '--type %r \\' % (semantic_type,)
        ]

        if view_type is not None:
            if type(view_type) is not str:
                view_type = view_type.__name__
            lines.append(self.INDENT + '--input-format %s \\' % (view_type,))

        lines += [
            self.INDENT + '--input-path %s \\' % (in_fp,),
            self.INDENT + '--output-path %s' % (out_fp,),
        ]

        self.recorder.extend(lines)

        return imported_var

    def init_format(self, name, factory, ext=None):
        if ext is not None:
            name = '%s.%s' % (name, ext.lstrip('.'))

        variable = super().init_format(name, factory, ext=ext)

        self.init_data.append(variable)

        return variable

    def init_metadata(self, name, factory):
        variable = super().init_metadata(name, factory)

        self.init_data.append(variable)

        return variable

    def comment(self, text):
        self.recorder += ['# ' + ln for ln in textwrap.wrap(text, width=74)]

    def peek(self, variable):
        var_name = variable.to_interface_name()
        self.recorder.append('qiime tools peek %s' % var_name)

    def merge_metadata(self, name, *variables):
        var = super().merge_metadata(name, *variables)

        # this is our special "short-circuit" attr to handle special-case
        # .to_interface_name() needs
        var._q2cli_ref = ' '.join(v.to_interface_name() for v in variables)
        return var

    def get_metadata_column(self, name, column_name, variable):
        var = super().get_metadata_column(name, column_name, variable)

        # this is our special "short-circuit" attr to handle special-case
        # .to_interface_name() needs
        var._q2cli_ref = (variable.to_interface_name(), column_name)
        return var

    def view_as_metadata(self, name, variable):
        # use the given name so that namespace behaves as expected,
        # then overwrite it because viewing is a no-op in q2cli
        var = super().view_as_metadata(name, variable)
        # preserve the original interface name of the QZA as this will be
        # implicitly converted to metadata when executed.
        var._q2cli_ref = variable.to_interface_name()
        return var

    def action(self, action, inputs, outputs):
        variables = super().action(action, inputs, outputs)

        vars_dict = variables._asdict()

        plugin_name = util.to_cli_name(action.plugin_id)
        action_name = util.to_cli_name(action.action_id)
        self.recorder.append('qiime %s %s \\' % (plugin_name, action_name))

        action_f = action.get_action()
        action_state = get_action_state(action_f)

        ins = inputs.map_variables(lambda v: v.to_interface_name())
        outs = {k: v.to_interface_name() for k, v in vars_dict.items()}
        signature = {s['name']: s for s in action_state['signature']}

        for param_name, value in ins.items():
            self._append_action_line(signature, param_name, value)

        max_collection_size = self.action_collection_size
        if max_collection_size is not None and len(outs) > max_collection_size:
            dir_name = self._build_output_dir_name(plugin_name, action_name)
            self.recorder.append(
                self.INDENT + '--output-dir %s \\' % (dir_name))
            self._rename_outputs(vars_dict, dir_name)
        else:
            for param_name, value in outs.items():
                self._append_action_line(signature, param_name, value)

        self.recorder[-1] = self.recorder[-1][:-2]  # remove trailing \

        return variables

    def _build_output_dir_name(self, plugin_name, action_name):
        base_name = '%s-%s' % (plugin_name, action_name)
        self.output_dir_counter[base_name] += 1
        current_inc = self.output_dir_counter[base_name]
        if current_inc == 1:
            return base_name
        return '%s-%d' % (base_name, current_inc)

    def _rename_outputs(self, vars_dict, dir_name):
        for signature_name, variable in vars_dict.items():
            name = '%s%s' % (signature_name, variable.ext)
            variable._q2cli_ref = os.path.join(dir_name, name)

    def _append_action_line(self, signature, param_name, value):
        param_state = signature[param_name]
        if value is not None:
            for opt, val in self._make_param(value, param_state):
                line = self.INDENT + opt
                if val is not None:
                    line += ' ' + val
                line += ' \\'

                self.recorder.append(line)

    def _make_param(self, value, state):
        state = state.copy()
        type_ = state.pop('type')

        opt = q2cli.click.option.GeneratedOption(prefix=type_[0], **state)
        option = opt.opts[0]

        # INPUTS AND OUTPUTS
        if type_ in ('input', 'output'):
            if isinstance(value, str):
                return [(option, value)]
            else:
                if isinstance(value, set):
                    value = sorted(value)
                return [(option, ' '.join(value))]

        # METADATA FILE
        if state['metadata'] == 'file':
            return [(option, value)]

        # METADATA COLUMN
        if state['metadata'] == 'column':
            # md cols are special, we have pre-computed the interface-specific
            # names and stashed them in an attr, so unpack to get the values
            fn, col_name = value
            return [(option, fn), (opt.q2_extra_opts[0], col_name)]

        # PARAMETERS
        if type_ == 'parameter':
            if isinstance(value, set):
                value = [shlex.quote(str(v)) for v in value]
                return [(option, ' '.join(sorted(value)))]

            if isinstance(value, list):
                return [(option, ' '.join(shlex.quote(str(v)) for v in value))]

            if isinstance(value, (dict, ResultCollection)):
                return [(option, ' '.join(f'{k}:{shlex.quote(str(v))}'
                                          for k, v in value.items()))]

            if type(value) is bool:
                if state['ast']['type'] == 'expression':
                    if value:
                        return [(option, None)]
                    else:
                        return [(opt.secondary_opts[0], None)]
                else:
                    # This is a more complicated param that can't be expressed
                    # as a typical `--p-foo/--p-no-foo` so default to baseline
                    # parameter handling behavior.
                    pass

            if type(value) is str:
                return [(option, shlex.quote(value))]

            return [(option, str(value))]

        raise Exception('Something went terribly wrong!')

    def get_example_data(self):
        for val in self.init_data:
            yield val.to_interface_name(), val.execute()


class ReplayCLIUsageVariable(CLIUsageVariable):
    def to_interface_name(self):
        '''
        Differs from parent method in that metadata is not kebab-cased,
        filepaths are preserved instead.
        '''
        if hasattr(self, '_q2cli_ref'):
            return self._q2cli_ref

        cli_name = '%s%s' % (self.name, self.ext)

        # don't disturb file names, this will break importing where QIIME 2
        # relies on specific filenames being present in a dir
        if self.var_type not in ('format', 'column', 'metadata'):
            cli_name = self.to_cli_name(cli_name)

        return cli_name


class ReplayCLIUsage(CLIUsage):
    shebang = '#!/usr/bin/env bash'
    header_boundary = ('#' * 79)
    set_ex = [
        '# This tells bash to -e exit immediately if a command fails',
        '# and -x show all commands in stdout so you can track progress',
        'set -e -x',
        ''
    ]
    copyright = pkg_resources.resource_string(
        __package__, 'assets/copyright_note.txt'
    ).decode('utf-8').split('\n')
    how_to = pkg_resources.resource_string(
        __package__, 'assets/cli_howto.txt'
    ).decode('utf-8').split('\n')

    def __init__(self, enable_assertions=False, action_collection_size=None):
        '''
        Identical to parent but creates header and footer attributes.

        Parameters
        ----------
        enable_assertions : bool
            Whether to render has-line-matching and output type assertions.
        action_collection_size : int
            The number of outputs returned by an action above which outputs are
            grouped into and accessed from an --output-dir.
        '''
        super().__init__()
        self.header = []
        self.footer = []
        self.enable_assertions = enable_assertions
        self.action_collection_size = action_collection_size

    def usage_variable(self, name, factory, var_type):
        return ReplayCLIUsageVariable(name, factory, var_type, self)

    def _append_action_line(self, signature, param_name: str, value):
        '''
        Extends the parent method to accommodate action signatures that may
        differ between those found in provenance and those accessible in the
        currently executing environment.

        Parameters
        ----------
        signature : dict of str -> dict
            Mapping of name of signature item to dict of signature spec data.
        param_name : str
            The name of the parameter for which to render a CLI line.
        value : any
            The value of an item from a Results.
        '''
        param_state = signature.get(param_name)
        if param_state is not None:
            for opt, val in self._make_param(value, param_state):
                line = self.INDENT + opt
                if val is not None:
                    line += ' ' + val
                line += ' \\'
                self.recorder.append(line)
        else:  # no matching param name
            line = self.INDENT + (
                '# FIXME: The following parameter name was not found in '
                'your current\n  # QIIME 2 environment. This may occur '
                'when the plugin version you have\n  # installed does not '
                'match the version used in the original analysis.\n  # '
                'Please see the docs and correct the parameter name '
                'before running.\n')
            cli_name = re.sub('_', '-', param_name)
            line += self.INDENT + '--?-' + cli_name + ' ' + str(value)
            line += ' \\'
            self.recorder.append(line)

    def _make_param(self, value: Any, state: Dict) -> List[Tuple]:
        '''
        Wraps metadata filenames in <> to force users to replace them.

        Parameters
        ----------
        value : any
            A value of a Result.
        state : dict
             A collection of info about an item from an action signature.
             See q2cli.core.state.py.

        Returns
        -------
        list of tuple
            See q2cli.core.usage.CLIUsage._make_param for possible outputs.
        '''
        if state['metadata'] == 'column':
            value = (f'{value[0]}', *value[1:])
        if state['metadata'] == 'file':
            value = f'{value}'
        return super()._make_param(value, state)

    def import_from_format(
        self,
        name: str,
        semantic_type: str,
        variable: UsageVariable,
        view_type: Any = None
    ) -> UsageVariable:
        '''
        Identical to super.import_from_format, but writes
        --input-path <yourdata here> and follows import block with a
        blank line.

        Parameters
        ----------
        name : str
            The name of the created UsageVariable.
        semantic_type : str
            The semantic type of the created UsageVariable.
        variable : UsageVariable
            A UsageVariable of some format type that will materialize the
            data to be imported.
        view_type : format or str
            The view type for importing.

        Returns
        -------
        UsageVariable
            Of type artifact.
        '''
        # need the super().super() here, so pass self to super
        imported_var = Usage.import_from_format(
            self, name, semantic_type, variable, view_type=view_type
        )

        out_fp = imported_var.to_interface_name()

        lines = [
            'qiime tools import \\',
            self.INDENT + '--type %r \\' % (semantic_type,)
        ]

        if view_type is not None:
            lines.append(
                self.INDENT + '--input-format %s \\' % (view_type,)
            )

        lines += [
            self.INDENT + '--input-path <your data here> \\',
            self.INDENT + '--output-path %s' % (out_fp,),
        ]

        lines.append('')
        self.recorder.extend(lines)

        return imported_var

    def init_metadata(
        self, name: str, factory: Callable, dumped_md_fn: str = ''
    ) -> UsageVariable:
        '''
        Like parent, but appropriately handles filepaths for recorded md fps.

        Parameters
        ----------
        name : str
            The name of the UsageVariable to be created.
        factory : Callable
            The factory responsible for generating the realized value of the
            UsageVariable.

        Returns
        -------
        UsageVariable
            Of type metadata.
        '''
        variable = super().init_metadata(name, factory)

        self.init_data.append(variable)

        if dumped_md_fn:
            variable.name = f'"{dumped_md_fn}.tsv"'
        else:
            variable.name = '<your metadata filepath>'

        return variable

    def comment(self, text):
        '''
        Identical to parent method, but pads comments with an extra newline.
        '''
        super().comment(text)
        self.recorder.append('')

    def action(
        self,
        action: Action,
        inputs: UsageInputs,
        outputs: UsageOutputNames
    ) -> UsageOutputs:
        '''
        Overrides parent method to fill in missing outputlines from
        action_f.signature.
        Also pads actions with an extra newline.

        Parameters
        ----------
        action : Action
            The underlying sdk.Action object.
        inputs : UsageInputs
            Mapping of parameter names to arguments for the action.
        outputs : UsageOutputNames
            Mapping of registered output names to usage variable names.

        Returns
        -------
        UsageOutputs
            The results returned by the action.
        '''
        variables = Usage.action(self, action, inputs, outputs)
        vars_dict = variables._asdict()

        # get registered collection of output names so we don't miss any
        action_f = action.get_action()
        missing_outputs = {}
        for output in action_f.signature.outputs:
            try:
                # If we get a match on output-name, the correct pair is already
                # in vars_dict and we can continue
                getattr(variables, output)
            except AttributeError:
                # Otherwise, we should add filler values to missing_outputs
                missing_outputs[output] = f'XX_{output}'

        plugin_name = q2cli.util.to_cli_name(action.plugin_id)
        action_name = q2cli.util.to_cli_name(action.action_id)
        self.recorder.append('qiime %s %s \\' % (plugin_name, action_name))

        action_f = action.get_action()
        action_state = get_action_state(action_f)

        ins = inputs.map_variables(lambda v: v.to_interface_name())
        outs = {k: v.to_interface_name() for k, v in vars_dict.items()}
        outs.update(missing_outputs)
        signature = {s['name']: s for s in action_state['signature']}

        for param_name, value in ins.items():
            self._append_action_line(signature, param_name, value)

        max_collection_size = self.action_collection_size
        if max_collection_size is not None and len(outs) > max_collection_size:
            dir_name = self._build_output_dir_name(plugin_name, action_name)
            self.recorder.append(
                self.INDENT + '--output-dir %s \\' % (dir_name)
            )
            self._rename_outputs(vars_dict, dir_name)
        else:
            for param_name, value in outs.items():
                self._append_action_line(signature, param_name, value)

        self.recorder[-1] = self.recorder[-1][:-2]  # remove trailing `\`

        self.recorder.append('')
        return variables

    def render(self, flush=False):
        '''
        Return a newline-seperated string of CLI script.

        Parameters
        ----------
        flush : bool
            Whether to 'flush' the current code. Importantly, this will clear
            the top-line imports for future invocations.

        Returns
        -------
        str
            The rendered string of CLI code.
        '''
        if self.header:
            self.header = self.header + ['']
        if self.footer:
            self.footer = [''] + self.footer
        rendered = '\n'.join(
            self.header + self.set_ex + self.recorder + self.footer
        )
        if flush:
            self.header = []
            self.footer = []
            self.recorder = []
            self.init_data = []
        return rendered

    def build_header(self):
        '''Constructs a renderable header from its components.'''
        self.header.extend(build_header(
            self.shebang, self.header_boundary, self.copyright, self.how_to
        ))

        # for creating result collections in bash
        bash_rc_function = [
            'construct_result_collection () {',
            '\tmkdir $rc_name',
            '\ttouch $rc_name.order',
            '\tfor key in "${keys[@]}"; do',
            '\t\techo $key >> $rc_name.order',
            '\tdone',
            '\tfor i in "${!keys[@]}"; do',
            '\t\tln -s ../"${names[i]}" $rc_name"${keys[i]}"$ext',
            '\tdone',
            '}'
        ]
        self.header.extend([
            '## function to create result collections ##',
            *bash_rc_function,
            '##',
        ])

    def build_footer(self, dag: ProvDAG):
        '''
        Constructs a renderable footer using the terminal uuids of a ProvDAG.
        '''
        self.footer.extend(build_footer(dag, self.header_boundary))



# ----------------------------------------------------------------------------
# Copyright (c) 2016-2023, QIIME 2 development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------


# NOTE: This module is closely coupled with `q2cli.cache`. It is a separate
# module to avoid cluttering `q2cli.cache` with completion script code, and
# provides a place to support other shell completion scripts in the future
# (e.g. zsh).

def write_bash_completion_script(plugins, path):
    """

    Parameters
    ----------
    plugins : dict
        Decoded JSON object representing CLI state on a per-plugin basis (e.g.
        as returned by `DeploymentCache.plugins`). See note within this
        function for why this parameter is necessary.
    path : str
        Path to write completion script to.

    """
    import os
    import os.path
    import stat
    import textwrap
    from q2cli.__main__ import qiime as root

    # `write_bash_completion_script` is called by `q2cli.cache.DeploymentCache`
    # when it is refreshing its cache. `q2cli.commands.RootCommand` could have
    # already asked for the cache, for example, if the user ran a command and
    # the cache must be refreshed. The bash completion script is generated by
    # traversing the `RootCommand` tree, so there is a cycle when `RootCommand`
    # attempts to access the cache in order to build itself. We work around
    # this by bootstrapping the `RootCommand`'s `._plugins` attribute with the
    # plugin state that has already been loaded by `DeploymentCache`.
    root._plugins = plugins

    cmd_reply = _generate_command_reply(root)
    cmd_reply = textwrap.indent(cmd_reply, '  ')
    completion_script = COMPLETION_SCRIPT_TEMPLATE.format(cmd_reply=cmd_reply)

    with open(path, 'w') as fh:
        fh.write(completion_script)

    # Make bash completion script executable:
    #   http://stackoverflow.com/a/12792002/3776794
    st = os.stat(path)
    # Set executable bit for user,group,other for root/sudo installs
    os.chmod(path, st.st_mode | stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH)


def _generate_command_reply(cmd):
    """Recursively generate completion reply for this command and subcommands.

    Parameters
    ----------
    cmd : click.Command
        Command to generate completion replies for (including its subcommands).

    """
    import textwrap
    import click

    ctx = None

    options = ['--help']
    for param in cmd.params:
        if isinstance(param, click.Option):
            options.extend(param.opts)
            options.extend(param.secondary_opts)
            if hasattr(param, 'q2_extra_opts'):
                options.extend(param.q2_extra_opts)

    subcmd_names = []
    if isinstance(cmd, click.MultiCommand):
        subcmd_names.extend(cmd.list_commands(ctx))

    subcmd_cases = []
    for subcmd_name in subcmd_names:
        subcmd_reply = _generate_command_reply(
            cmd.get_command(ctx, subcmd_name))
        subcmd_reply = textwrap.indent(subcmd_reply, '  ')

        case = SUBCOMMAND_CASE_TEMPLATE.format(
            subcmd_name=subcmd_name, subcmd_reply=subcmd_reply)
        subcmd_cases.append(case)

    subcmd_cases = textwrap.indent('\n'.join(subcmd_cases), ' ' * 6)

    cmd_reply = COMMAND_REPLY_TEMPLATE.format(
        options=' '.join(options), subcmd_names=' '.join(subcmd_names),
        subcmd_cases=subcmd_cases)

    return cmd_reply


# NOTE: using double braces to avoid `str.format` interpolation when bash needs
# curly braces in the generated code.
#
# NOTE: the handling of a negative COMP_CWORD is necessary in certain versions
# of bash (e.g. at least the bash shipped with OS X 10.9.5). When adding
# whitespace to the end of a command, and then moving the cursor backwards in
# the command and hitting <tab>, COMP_CWORD can be negative (I've only seen -2
# as its value). This is a bash bug and is not documented behavior. Other CLIs
# with tab completion suffer from the same issue, and each one deals with this
# bug differently (some not at all, e.g. `git`). The workaround used below
# seems to provide the least destructive completion behavior for our CLI.
#
# Bug report reference:
#   https://lists.gnu.org/archive/html/bug-bash/2009-07/msg00108.html

COMPLETION_SCRIPT_TEMPLATE = """\
#!/usr/bin/env bash

_qiime_completion()
{{
  local COMP_WORDS=(${{COMP_WORDS[*]}})
  local incomplete
  if [[ ${{COMP_CWORD}} -lt 0 ]] ; then
    COMP_CWORD="${{#COMP_WORDS[*]}}"
    incomplete=""
  else
    incomplete="${{COMP_WORDS[COMP_CWORD]}}"
  fi

  local curpos nextpos nextword
  nextpos=0

{cmd_reply}

  return 0
}}

_qiime_completion
"""

COMMAND_REPLY_TEMPLATE = """\
curpos=${{nextpos}}
while :
do
  nextpos=$((curpos + 1))
  nextword="${{COMP_WORDS[nextpos]}}"
  if [[ ${{nextpos}} -eq ${{COMP_CWORD}} ]] ; then
    if [[ ${{incomplete}} == -* ]] ; then
      echo "$(compgen -W "{options}" -- $incomplete)"
    else
      echo "$(compgen -W "{subcmd_names}" -- $incomplete)"
    fi
    return 0
  else
    case "${{nextword}}" in
{subcmd_cases}
    esac
    curpos=${{nextpos}}
  fi
done\
"""

SUBCOMMAND_CASE_TEMPLATE = """\
{subcmd_name})
{subcmd_reply}
  ;;
"""



# ----------------------------------------------------------------------------
# Copyright (c) 2016-2023, QIIME 2 development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

import os
import shutil
import unittest
import tempfile
import configparser

from click.testing import CliRunner
from qiime2 import Artifact
from qiime2.core.testing.type import IntSequence1
from qiime2.core.testing.util import get_dummy_plugin

import q2cli.util
from q2cli.builtin.dev import dev


class TestDev(unittest.TestCase):
    path = os.path.join(q2cli.util.get_app_dir(), 'cli-colors.theme')
    old_settings = None
    if os.path.exists(path):
        old_settings = configparser.ConfigParser()
        old_settings.read(path)

    def setUp(self):
        get_dummy_plugin()
        self.parser = configparser.ConfigParser()
        self.runner = CliRunner(mix_stderr=False)
        self.tempdir = tempfile.mkdtemp(prefix='qiime2-q2cli-test-temp-')
        self.generated_config = os.path.join(self.tempdir, 'generated-theme')

        self.tempdir = tempfile.mkdtemp(prefix='qiime2-q2cli-test-temp-')
        self.artifact1_path = os.path.join(self.tempdir, 'a1.qza')
        self.mapping_path = os.path.join(self.tempdir, 'mapping.qza')

        artifact1 = Artifact.import_data(IntSequence1, [0, 42, 43])
        artifact1.save(self.artifact1_path)
        self.artifact1_root_dir = str(artifact1.uuid)

        mapping = Artifact.import_data('Mapping', {'foo': '42'})
        mapping.save(self.mapping_path)

        self.config = os.path.join(self.tempdir, 'good-config.ini')
        self.parser['type'] = {'underline': 't'}
        with open(self.config, 'w') as fh:
            self.parser.write(fh)

    def tearDown(self):
        if self.old_settings is not None:
            with open(self.path, 'w') as fh:
                self.old_settings.write(fh)
        shutil.rmtree(self.tempdir)

    def test_import_theme(self):
        result = self.runner.invoke(
            dev, ['import-theme', '--theme', self.config])
        self.assertEqual(result.exit_code, 0)

    def test_export_default_theme(self):
        result = self.runner.invoke(
            dev, ['export-default-theme', '--output-path',
                  self.generated_config])
        self.assertEqual(result.exit_code, 0)

    def test_reset_theme(self):
        result = self.runner.invoke(
            dev, ['reset-theme', '--yes'])
        self.assertEqual(result.exit_code, 0)

    def test_reset_theme_no_yes(self):
        result = self.runner.invoke(
            dev, ['reset-theme'])
        self.assertNotEqual(result.exit_code, 0)

    # result_type & result_data tests
    def test_assert_result_type_artifact_success(self):
        result = self.runner.invoke(dev,
                                    ['assert-result-type',
                                     self.mapping_path,
                                     '--qiime-type', 'Mapping'])

        # single regex to account for tempdir path

        expected_regex = r'The input file \(.*mapping.qza\) type and the'\
                         r' expected type \(Mapping\) match'

        self.assertEqual(result.exit_code, 0)
        self.assertRegex(result.stdout, expected_regex)

    def test_assert_result_type_visualization_success(self):
        dummy_plugin = get_dummy_plugin()

        self.viz_path = os.path.join(self.tempdir, 'viz.qzv')
        most_common_viz = dummy_plugin.actions['most_common_viz']
        viz = most_common_viz(Artifact.load(self.artifact1_path))
        viz.visualization.save(self.viz_path)

        result = self.runner.invoke(dev,
                                    ['assert-result-type',
                                     self.viz_path,
                                     '--qiime-type',
                                     'Visualization'])

        expected_regex = r'The input file \(.*viz\.qzv\) type and the'\
                         r' expected type \(Visualization\) match'

        self.assertEqual(result.exit_code, 0)
        self.assertRegex(result.stdout, expected_regex)

    def test_assert_result_type_load_failure(self):
        result = self.runner.invoke(dev,
                                    ['assert-result-type',
                                     'turkey_sandwhere.qza',
                                     '--qiime-type', 'Mapping'])

        self.assertEqual(result.exit_code, 1)
        self.assertRegex(result.stderr,
                         r'File\s*\'turkey_sandwhere\.qza\'\s*does not exist')

    def test_assert_result_type_invalid_qiime_type(self):
        result = self.runner.invoke(dev,
                                    ['assert-result-type',
                                     self.mapping_path,
                                     '--qiime-type', 'Squid'])
        self.assertEqual(result.exit_code, 1)
        self.assertIn('Expected Squid, observed Mapping', result.stderr)

    def test_assert_result_data_success(self):
        result = self.runner.invoke(dev,
                                    ['assert-result-data',
                                     self.mapping_path,
                                     '--zip-data-path', 'mapping.tsv',
                                     '--expression', '42'])
        self.assertEqual(result.exit_code, 0)
        self.assertIn(r'"42" was found in mapping.tsv', result.stdout)

    def test_assert_result_data_load_failure(self):
        result = self.runner.invoke(dev,
                                    ['assert-result-data',
                                     'turkey_sandwhen.qza',
                                     '--zip-data-path', 'mapping.tsv',
                                     '--expression', '42'])

        self.assertEqual(result.exit_code, 1)
        self.assertRegex(result.stderr,
                         r'File\s*\'turkey_sandwhen\.qza\'\s*does not exist')

    def test_assert_result_data_zip_data_path_zero_matches(self):
        result = self.runner.invoke(dev,
                                    ['assert-result-data',
                                     self.mapping_path,
                                     '--zip-data-path', 'turkey_sandwhy.tsv',
                                     '--expression', '42'])

        self.assertEqual(result.exit_code, 1)
        self.assertRegex(result.stderr,
                         r'did not produce exactly one match.\n'
                         r'Matches: \[\]\n')

    def test_assert_result_data_zip_data_path_multiple_matches(self):
        self.double_path = os.path.join(self.tempdir, 'double.qza')
        double_artifact = Artifact.import_data('SingleInt', 3)
        double_artifact.save(self.double_path)
        result = self.runner.invoke(dev, ['assert-result-data',
                                          self.double_path,
                                          '--zip-data-path',
                                          'file*.txt',
                                          '--expression',
                                          '3'])
        self.assertEqual(result.exit_code, 1)
        self.assertRegex(result.stderr, r'Value provided for zip_data_path'
                                        r' \(file\*\.txt\) did not produce'
                                        r' exactly one match\.')

    def test_assert_result_data_match_expression_not_found(self):
        result = self.runner.invoke(dev, ['assert-result-data',
                                          self.mapping_path,
                                          '--zip-data-path', 'mapping.tsv',
                                          '--expression', 'foobar'])

        self.assertEqual(result.exit_code, 1)
        self.assertRegex(result.stderr,
                         r'Expression \'foobar\''
                         r' not found in .*\/data\/mapping\.tsv\.')



# ----------------------------------------------------------------------------
# Copyright (c) 2016-2023, QIIME 2 development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

import os.path
import pathlib
import shutil
import tempfile
import unittest
import configparser
import zipfile

import pandas as pd

from click.testing import CliRunner
from qiime2 import Artifact
from qiime2.core.testing.type import IntSequence1
from qiime2.core.testing.util import get_dummy_plugin
from qiime2.sdk.util import camel_to_snake
from qiime2.sdk.usage import UsageVariable
from qiime2.sdk import PluginManager
from qiime2.core.archive.provenance_lib import DummyArtifacts, ProvDAG
from qiime2.core.archive.provenance_lib.replay import (
    ReplayConfig, param_is_metadata_column, dump_recorded_md_file,
    ReplayNamespaces, build_import_usage, build_action_usage,
    ActionCollections, replay_provenance, replay_supplement
)
from qiime2.core.archive.provenance_lib.usage_drivers import ReplayPythonUsage

import q2cli
import q2cli.util
import q2cli.builtin.info
import q2cli.builtin.tools
from q2cli.commands import RootCommand
from q2cli.core.config import CLIConfig
from q2cli.core.usage import ReplayCLIUsage, CLIUsageVariable


class TestOption(unittest.TestCase):
    def setUp(self):
        get_dummy_plugin()
        self.runner = CliRunner()
        self.tempdir = tempfile.mkdtemp(prefix='qiime2-q2cli-test-temp-')

        self.parser = configparser.ConfigParser()
        self.path = os.path.join(q2cli.util.get_app_dir(), 'cli-colors.theme')

    def tearDown(self):
        shutil.rmtree(self.tempdir)

    def _assertRepeatedOptionError(self, result, option):
        self.assertEqual(result.exit_code, 1)
        self.assertTrue(result.output.startswith('Usage:'))
        self.assertRegex(result.output, '.*%s.* was specified multiple times'
                         % option)

    def test_repeated_eager_option_with_callback(self):
        result = self.runner.invoke(
            q2cli.builtin.tools.tools,
            ['list-types', '--tsv', '--tsv'])

        self._assertRepeatedOptionError(result, '--tsv')

    def test_repeated_builtin_flag(self):
        result = self.runner.invoke(
            q2cli.builtin.tools.tools,
            ['import', '--input-path', 'a', '--input-path', 'b'])

        self._assertRepeatedOptionError(result, '--input-path')

    def test_repeated_action_flag(self):
        qiime_cli = RootCommand()
        command = qiime_cli.get_command(ctx=None, name='dummy-plugin')
        out_path = os.path.join(self.tempdir, 'out.qza')

        result = self.runner.invoke(
            command, ['no-input-method', '--o-out', out_path,
                      '--verbose', '--verbose'])

        self._assertRepeatedOptionError(result, '--verbose')

    def test_repeated_builtin_option(self):
        input_path = os.path.join(self.tempdir, 'ints.txt')
        with open(input_path, 'w') as f:
            f.write('42\n43\n44\n')
        output_path = os.path.join(self.tempdir, 'out.qza')

        result = self.runner.invoke(
            q2cli.builtin.tools.tools,
            ['import', '--input-path', input_path,
             '--output-path', output_path, '--type', 'IntSequence1',
             '--type', 'IntSequence1'])

        self._assertRepeatedOptionError(result, '--type')

    def test_repeated_action_option(self):
        qiime_cli = RootCommand()
        command = qiime_cli.get_command(ctx=None, name='dummy-plugin')
        out_path = os.path.join(self.tempdir, 'out.qza')

        result = self.runner.invoke(
            command, ['no-input-method', '--o-out', out_path,
                      '--o-out', out_path])

        self._assertRepeatedOptionError(result, '--o-out')

    def test_repeated_multiple_option(self):
        input_path = os.path.join(self.tempdir, 'ints.qza')
        artifact = Artifact.import_data(IntSequence1, [0, 42, 43], list)
        artifact.save(input_path)

        metadata_path1 = os.path.join(self.tempdir, 'metadata1.tsv')
        with open(metadata_path1, 'w') as f:
            f.write('id\tcol1\nid1\tfoo\nid2\tbar\n')
        metadata_path2 = os.path.join(self.tempdir, 'metadata2.tsv')
        with open(metadata_path2, 'w') as f:
            f.write('id\tcol2\nid1\tbaz\nid2\tbaa\n')

        output_path = os.path.join(self.tempdir, 'out.qza')

        qiime_cli = RootCommand()
        command = qiime_cli.get_command(ctx=None, name='dummy-plugin')

        result = self.runner.invoke(
            command, ['identity-with-metadata', '--i-ints', input_path,
                      '--o-out', output_path, '--m-metadata-file',
                      metadata_path1, '--m-metadata-file', metadata_path2,
                      '--verbose'])

        self.assertEqual(result.exit_code, 0)
        self.assertTrue(os.path.exists(output_path))
        self.assertEqual(Artifact.load(output_path).view(list), [0, 42, 43])

    def test_config_expected(self):
        self.parser['type'] = {'underline': 't'}
        with open(self.path, 'w') as fh:
            self.parser.write(fh)

        config = CLIConfig()
        config.parse_file(self.path)

        self.assertEqual(
            config.styles['type'], {'underline': True})

    def test_config_bad_selector(self):
        self.parser['tye'] = {'underline': 't'}
        with open(self.path, 'w') as fh:
            self.parser.write(fh)

        config = CLIConfig()
        with self.assertRaisesRegex(
                configparser.Error, 'tye.*valid selector.*valid selectors'):
            config.parse_file(self.path)

    def test_config_bad_styling(self):
        self.parser['type'] = {'underlined': 't'}
        with open(self.path, 'w') as fh:
            self.parser.write(fh)

        config = CLIConfig()
        with self.assertRaisesRegex(
                configparser.Error, 'underlined.*valid styling.*valid '
                'stylings'):
            config.parse_file(self.path)

    def test_config_bad_color(self):
        self.parser['type'] = {'fg': 'purple'}
        with open(self.path, 'w') as fh:
            self.parser.write(fh)

        config = CLIConfig()
        with self.assertRaisesRegex(
                configparser.Error, 'purple.*valid color.*valid colors'):
            config.parse_file(self.path)

    def test_config_bad_boolean(self):
        self.parser['type'] = {'underline': 'g'}
        with open(self.path, 'w') as fh:
            self.parser.write(fh)

        config = CLIConfig()
        with self.assertRaisesRegex(
                configparser.Error, 'g.*valid boolean.*valid booleans'):
            config.parse_file(self.path)

    def test_no_file(self):
        config = CLIConfig()
        with self.assertRaisesRegex(
                configparser.Error, "'Path' is not a valid filepath."):
            config.parse_file('Path')


class ReplayCLIUsageTests(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.das = DummyArtifacts()
        cls.tempdir = cls.das.tempdir
        cls.pm = PluginManager()

    @classmethod
    def tearDownClass(cls):
        cls.das.free()

    def test_init_metadata(self):
        use = ReplayCLIUsage()
        var = use.init_metadata(name='testing', factory=lambda: None)
        self.assertEqual(var.name, '<your metadata filepath>')
        self.assertEqual(var.var_type, 'metadata')

    def test_init_metadata_with_dumped_md_fn(self):
        use = ReplayCLIUsage()
        var = use.init_metadata(
            name='testing', factory=lambda: None, dumped_md_fn='some_md')
        self.assertEqual(var.var_type, 'metadata')
        self.assertEqual(var.name, '"some_md.tsv"')

    def test_param_is_metadata_col(self):
        cfg = ReplayConfig(use=ReplayCLIUsage(),
                           use_recorded_metadata=False, pm=self.pm)

        actual = param_is_metadata_column(
            cfg, 'metadata', 'dummy_plugin', 'identity_with_metadata_column'
        )
        self.assertTrue(actual)

        actual = param_is_metadata_column(
            cfg, 'int1', 'dummy_plugin', 'concatenate_ints'
        )
        self.assertFalse(actual)

        with self.assertRaisesRegex(KeyError, "No action.*registered.*"):
            param_is_metadata_column(
                cfg, 'ints', 'dummy_plugin', 'young'
            )

        with self.assertRaisesRegex(KeyError, "No param.*registered.*"):
            param_is_metadata_column(
                cfg, 'thugger', 'dummy_plugin', 'split_ints'
            )

        with self.assertRaisesRegex(KeyError, "No plugin.*registered.*"):
            param_is_metadata_column(
                cfg, 'fake_param', 'dummy_hard', 'split_ints'
            )

    def test_dump_recorded_md_file_to_custom_dir(self):
        dag = self.das.int_seq_with_md.dag
        uuid = self.das.int_seq_with_md.uuid

        out_dir = 'custom_dir'
        provnode = dag.get_node_data(uuid)
        og_md = provnode.metadata['metadata']
        action_name = 'concatenate_ints_0'
        md_id = 'metadata'
        fn = 'metadata.tsv'

        with tempfile.TemporaryDirectory() as tempdir:
            cfg = ReplayConfig(use=ReplayCLIUsage(),
                               pm=self.pm,
                               md_out_dir=(tempdir + '/' + out_dir))
            dump_recorded_md_file(cfg, provnode, action_name, md_id, fn)
            out_path = pathlib.Path(tempdir) / out_dir / action_name / fn

            self.assertTrue(out_path.is_file())

            dumped_df = pd.read_csv(out_path, sep='\t')
            pd.testing.assert_frame_equal(dumped_df, og_md)

            # If we run it again, it shouldn't overwrite 'recorded_metadata',
            # so we should have two files
            action_name_2 = 'concatenate_ints_1'
            md_id2 = 'metadata'
            fn2 = 'metadata_1.tsv'
            dump_recorded_md_file(cfg, provnode, action_name_2, md_id2, fn2)
            out_path2 = pathlib.Path(tempdir) / out_dir / action_name_2 / fn2

            # are both files where expected?
            self.assertTrue(out_path.is_file())
            self.assertTrue(out_path2.is_file())

    def test_build_import_usage_cli(self):
        ns = ReplayNamespaces()
        cfg = ReplayConfig(use=ReplayCLIUsage(),
                           use_recorded_metadata=False, pm=self.pm)
        dag = self.das.concated_ints_v6.dag
        import_uuid = '8dea2f1a-2164-4a85-9f7d-e0641b1db22b'
        import_node = dag.get_node_data(import_uuid)
        c_to_s_type = camel_to_snake(import_node.type)
        unq_var_nm = c_to_s_type + '_0'
        build_import_usage(import_node, ns, cfg)
        rendered = cfg.use.render()
        usg_var = ns.get_usg_var_record(import_uuid).variable
        out_name = usg_var.to_interface_name()

        self.assertIsInstance(usg_var, UsageVariable)
        self.assertEqual(usg_var.var_type, 'artifact')
        self.assertEqual(usg_var.name, unq_var_nm)
        self.assertRegex(rendered, r'qiime tools import \\')
        self.assertRegex(rendered, f"  --type '{import_node.type}'")
        self.assertRegex(rendered, "  --input-path <your data here>")
        self.assertRegex(rendered, f"  --output-path {out_name}")

    def test_build_action_usage_cli(self):
        plugin = 'dummy-plugin'
        action = 'concatenate-ints'
        cfg = ReplayConfig(use=ReplayCLIUsage(),
                           use_recorded_metadata=False, pm=self.pm)

        ns = ReplayNamespaces()
        import_var_1 = CLIUsageVariable(
            'imported_ints_0', lambda: None, 'artifact', cfg.use
        )
        import_var_2 = CLIUsageVariable(
            'imported_ints_1', lambda: None, 'artifact', cfg.use
        )
        import_uuid_1 = '8dea2f1a-2164-4a85-9f7d-e0641b1db22b'
        import_uuid_2 = '7727c060-5384-445d-b007-b64b41a090ee'
        ns.add_usg_var_record(import_uuid_1, 'imported_ints_0')
        ns.update_usg_var_record(import_uuid_1, import_var_1)
        ns.add_usg_var_record(import_uuid_2, 'imported_ints_1')
        ns.update_usg_var_record(import_uuid_2, import_var_2)

        dag = self.das.concated_ints_v6.dag
        action_uuid = '5035a60e-6f9a-40d4-b412-48ae52255bb5'
        node_uuid = '6facaf61-1676-45eb-ada0-d530be678b27'
        node = dag.get_node_data(node_uuid)
        actions = ActionCollections(
            std_actions={action_uuid: {node_uuid: 'concatenated_ints'}}
        )
        unique_var_name = node.action.output_name + '_0'
        build_action_usage(node, ns, actions.std_actions, action_uuid, cfg)
        rendered = cfg.use.render()

        usg_var = ns.get_usg_var_record(node_uuid).variable
        out_name = usg_var.to_interface_name()

        self.assertIsInstance(usg_var, UsageVariable)
        self.assertEqual(usg_var.var_type, 'artifact')
        self.assertEqual(usg_var.name, unique_var_name)

        self.assertIn(f'qiime {plugin} {action}', rendered)
        self.assertIn('--i-ints1 imported-ints-0.qza', rendered)
        self.assertIn('--i-ints3 imported-ints-1.qza', rendered)
        self.assertIn('--p-int1 7', rendered)
        self.assertIn(f'--o-concatenated-ints {out_name}', rendered)

    def test_replay_optional_param_is_none(self):
        dag = self.das.int_seq_optional_input.dag
        with tempfile.TemporaryDirectory() as tempdir:
            out_path = pathlib.Path(tempdir) / 'ns_coll.txt'
            replay_provenance(ReplayCLIUsage, dag, out_path,
                              md_out_dir=tempdir)

            with open(out_path, 'r') as fp:
                rendered = fp.read()
            self.assertIn('--i-ints int-sequence1-0.qza', rendered)
            self.assertIn('--p-num1', rendered)
            self.assertNotIn('--i-optional1', rendered)
            self.assertNotIn('--p-num2', rendered)

    def test_replay_from_provdag_ns_collision(self):
        """
        This artifact's dag contains a few results with the output-name
        filtered-table, so is a good check for namespace collisions if
        we're not uniquifying variable names properly.
        """
        with tempfile.TemporaryDirectory() as tempdir:
            self.das.concated_ints.artifact.save(
                os.path.join(tempdir, 'c1.qza')
            )
            self.das.other_concated_ints.artifact.save(
                os.path.join(tempdir, 'c2.qza')
            )
            dag = ProvDAG(tempdir)

        exp = ['concatenated-ints-0', 'concatenated-ints-1']
        with tempfile.TemporaryDirectory() as tempdir:
            out_path = pathlib.Path(tempdir) / 'ns_coll.txt'
            replay_provenance(ReplayCLIUsage, dag, out_path,
                              md_out_dir=tempdir)

            with open(out_path, 'r') as fp:
                rendered = fp.read()
                for name in exp:
                    self.assertIn(name, rendered)


class WriteReproducibilitySupplementTests(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.das = DummyArtifacts()
        cls.tempdir = cls.das.tempdir
        cls.pm = PluginManager()

    @classmethod
    def tearDownClass(cls):
        cls.das.free()

    def test_replay_supplement_from_fp(self):
        fp = self.das.concated_ints_with_md.filepath
        with tempfile.TemporaryDirectory() as tempdir:
            out_fp = os.path.join(tempdir, 'supplement.zip')
            replay_supplement(
                usage_drivers=[ReplayPythonUsage, ReplayCLIUsage],
                payload=fp,
                out_fp=out_fp
            )

            self.assertTrue(zipfile.is_zipfile(out_fp))

            exp = {
                'supplement/',
                'supplement/python3_replay.py',
                'supplement/cli_replay.sh',
                'supplement/citations.bib',
                'supplement/recorded_metadata/',
                'supplement/recorded_metadata/'
                'dummy_plugin_identity_with_metadata_0/metadata_0.tsv',
            }
            with zipfile.ZipFile(out_fp, 'r') as myzip:
                namelist_set = set(myzip.namelist())
                for item in exp:
                    self.assertIn(item, namelist_set)

    def test_replay_supplement_from_provdag(self):
        dag = self.das.concated_ints_with_md.dag

        with tempfile.TemporaryDirectory() as tempdir:
            out_fp = os.path.join(tempdir, 'supplement.zip')
            replay_supplement(
                usage_drivers=[ReplayPythonUsage, ReplayCLIUsage],
                payload=dag,
                out_fp=out_fp
            )

            self.assertTrue(zipfile.is_zipfile(out_fp))

            exp = {
                'supplement/',
                'supplement/python3_replay.py',
                'supplement/cli_replay.sh',
                'supplement/citations.bib',
                'supplement/recorded_metadata/',
                'supplement/recorded_metadata/'
                'dummy_plugin_identity_with_metadata_0/metadata_0.tsv',
            }
            with zipfile.ZipFile(out_fp, 'r') as myzip:
                namelist_set = set(myzip.namelist())
                for item in exp:
                    self.assertIn(item, namelist_set)


if __name__ == "__main__":
    unittest.main()



# ----------------------------------------------------------------------------
# Copyright (c) 2016-2023, QIIME 2 development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

import click

from .type import QIIME2Type

# Sentinel to avoid the situation where `None` *is* the default value.
NoDefault = {}


class GeneratedOption(click.Option):
    def __init__(self, *, prefix, name, repr, ast, multiple, is_bool_flag,
                 metadata, metavar, default=NoDefault, description=None,
                 **attrs):
        import q2cli.util

        if metadata is not None:
            prefix = 'm'
        if multiple is not None:
            if multiple == 'list':
                multiple = list
            elif multiple == 'dict':
                multiple = dict
            else:
                multiple = set

        if is_bool_flag:
            yes = q2cli.util.to_cli_name(name)
            no = q2cli.util.to_cli_name('no_' + name)
            opt = f'--{prefix}-{yes}/--{prefix}-{no}'
        elif metadata is not None:
            cli_name = q2cli.util.to_cli_name(name)
            opt = f'--{prefix}-{cli_name}-file'
            if metadata == 'column':
                self.q2_extra_dest, self.q2_extra_opts, _ = \
                    self._parse_decls([f'--{prefix}-{cli_name}-column'], True)
        else:
            cli_name = q2cli.util.to_cli_name(name)
            opt = f'--{prefix}-{cli_name}'

        click_type = QIIME2Type(ast, repr, is_output=prefix == 'o')
        attrs['metavar'] = metavar
        attrs['multiple'] = multiple is not None
        attrs['param_decls'] = [opt]
        attrs['required'] = default is NoDefault
        attrs['help'] = self._add_default(description, default)
        if default is not NoDefault:
            attrs['default'] = default

        # This is to evade clicks __DEBUG__ check
        if not is_bool_flag:
            attrs['type'] = click_type
        else:
            attrs['type'] = None

        # This nonsense:
        # https://github.com/pallets/click/blob
        # /08f71b08e2b7ee9b1ea27daf6d3040999fc68551
        # /src/click/core.py#L2576-L2584
        if is_bool_flag and multiple is not None:
            to_add_multiple = attrs.pop('multiple')

        super().__init__(**attrs)

        if is_bool_flag and multiple is not None:
            self.multiple = to_add_multiple

        # put things back the way they _should_ be after evading __DEBUG__
        self.is_bool_flag = is_bool_flag
        self.type = click_type

        # attrs we will use elsewhere
        self.q2_multiple = multiple
        self.q2_prefix = prefix
        self.q2_name = name
        self.q2_ast = ast
        self.q2_metadata = metadata

    @property
    def meta_help(self):
        if self.q2_metadata == 'file':
            return 'multiple arguments will be merged'

    def _add_default(self, desc, default):
        if desc is not None:
            desc += '  '
        else:
            desc = ''
        if default is not NoDefault:
            if default is None:
                desc += '[optional]'
            else:
                desc += '[default: %r]' % (default,)
        return desc

    def consume_value(self, ctx, opts):
        if self.q2_metadata == 'column':
            return self._consume_metadata(ctx, opts)
        else:
            return super().consume_value(ctx, opts)

    def _consume_metadata(self, ctx, opts):
        # double consume
        # this consume deals with the metadata file
        md_file, source = super().consume_value(ctx, opts)
        # consume uses self.name, so mutate but backup for after
        backup, self.name = self.name, self.q2_extra_dest
        try:
            # this consume deals with the metadata column
            md_col, _ = super().consume_value(ctx, opts)
        # If `--m-metadata-column` isn't provided, need to set md_col to None
        # in order for the click.MissingParameter errors below to be raised
        except click.MissingParameter:
            md_col = None

        self.name = backup
        # These branches won't get hit unless there's a value associated with
        # md_col - the try/except case above handled the situation where the
        # metadata_column parameter itself wasn't provided (vs just a value)
        if (md_col is None) != (md_file is None):
            # missing one or the other
            if md_file is None:
                raise click.MissingParameter(ctx=ctx, param=self)
            else:
                raise click.MissingParameter(param_hint=self.q2_extra_opts,
                                             ctx=ctx, param=self)

        if md_col is None and md_file is None:
            return (None, source)
        else:
            return ((md_file, md_col), source)

    def get_help_record(self, ctx):
        record = super().get_help_record(ctx)
        if self.is_bool_flag:
            metavar = self.make_metavar()
            if metavar:
                record = (record[0] + ' ' + self.make_metavar(), record[1])
        elif self.q2_metadata == 'column':
            opts = (record[0], self.q2_extra_opts[0] + ' COLUMN ')
            record = (opts, record[1])
        return record

    # Override
    def add_to_parser(self, parser, ctx):
        shared = dict(dest=self.name, nargs=0, obj=self)
        if self.q2_metadata == 'column':
            parser.add_option(opts=self.opts, action='store', dest=self.name,
                              nargs=1, obj=self)
            parser.add_option(opts=self.q2_extra_opts, action='store',
                              dest=self.q2_extra_dest, nargs=1, obj=self)
        elif self.is_bool_flag:
            if self.multiple:
                action = 'append_maybe'
            else:
                action = 'store_maybe'
            parser.add_option(opts=self.opts, action=action, const=True,
                              **shared)
            parser.add_option(opts=self.secondary_opts, action=action,
                              const=False, **shared)
        elif self.multiple:
            action = 'append_greedy'
            parser.add_option(opts=self.opts, action='append_greedy', **shared)
        else:
            super().add_to_parser(parser, ctx)

    def get_default(self, ctx, call=True):
        if self.required and not ctx.resilient_parsing and not (
                self.q2_prefix == 'o' and ctx.params.get('output_dir', False)):
            raise click.MissingParameter(ctx=ctx, param=self)
        return super().get_default(ctx, call=call)

    def process_value(self, ctx, value):
        try:
            return super().process_value(ctx, value)
        except click.MissingParameter:
            if not (self.q2_prefix == 'o'
                    and ctx.params.get('output_dir', False)):
                raise

    def type_cast_value(self, ctx, value):
        import sys
        import q2cli.util
        import qiime2.sdk.util

        if self.multiple:
            if value == () or value is None:
                return None
            elif self.q2_prefix == 'i':
                value = super().type_cast_value(ctx, value)
                keys, value = self._split_and_validate_input_keys(value)

                if self.q2_multiple is set:
                    self._check_length(value, ctx)

                # This means we loaded a proper Collection directory. When we
                # load in a Collection directory for an action that takes a
                # Collection input, we get a tuple containing a dictionary of
                # the Collection we wanted. When we load in a Collection
                # directory for an action that takes a List, we get a list
                # containing a dictionary of the Collection we wanted. We just
                # extract that dictionary.
                if (isinstance(value, tuple) or isinstance(value, list)) \
                        and len(value) == 1 and isinstance(value[0], dict):
                    value = value[0]

                # We already have a dict, so we already have keys
                if isinstance(value, dict):
                    keys = value.keys()
                    value = list(value.values())
                elif self.q2_multiple is dict:
                    if keys is None:
                        keys = range(len(value))
                    value = value
                else:
                    value = self.q2_multiple(value)

                type_expr = qiime2.sdk.util.type_from_ast(self.q2_ast)
                args = ', '.join(map(repr, (x.type for x in value)))

                if value not in type_expr:
                    raise click.BadParameter(
                        'received <%s> as an argument, which is incompatible'
                        ' with parameter type: %r' % (args, type_expr),
                        ctx=ctx, param=self)

                if self.q2_multiple is dict:
                    value = {str(k): v for k, v in zip(keys, value)}
                return value
            elif self.q2_metadata == 'file':
                value = super().type_cast_value(ctx, value)
                if len(value) == 1:
                    return value[0]
                else:
                    try:
                        return value[0].merge(*value[1:])
                    except Exception as e:
                        header = ("There was an issue with merging "
                                  "QIIME 2 Metadata:")
                        tb = 'stderr' if '--verbose' in sys.argv else None
                        q2cli.util.exit_with_error(
                            e, header=header, traceback=tb)
            elif self.q2_prefix == 'p':
                try:
                    _values = []

                    if self.q2_multiple is set:
                        self._check_length(value, ctx)

                    keys = []
                    if self.q2_multiple is dict and type(value) is not dict:
                        _values = {}

                        keyed = False
                        unkeyed = False
                        # All params in a Collection must be either keyed or
                        # unkeyed. We cannot have a mix because it makes things
                        # ambiguous
                        for idx, item in enumerate(value):
                            if ':' in item:
                                if unkeyed:
                                    raise KeyError(
                                        'The keyed value <%s> has been mixed'
                                        ' with unkeyed values. All values must'
                                        ' be keyed or unkeyed.' % item)
                                key, _value = item.split(':', 1)
                                _values[key] = _value
                                keyed = True
                            else:
                                if keyed:
                                    raise KeyError(
                                        'The unkeyed value <%s> has been'
                                        ' mixed with keyed values. All values'
                                        ' must be keyed or unkeyed.' % item)
                                _values[str(idx)] = item
                                unkeyed = True
                    else:
                        _values = value

                    value = \
                        qiime2.sdk.util.parse_primitive(self.q2_ast, _values)
                except ValueError:
                    args = ', '.join(map(repr, value))
                    expr = qiime2.sdk.util.type_from_ast(self.q2_ast)
                    raise click.BadParameter(
                        'received <%s> as an argument, which is incompatible'
                        ' with parameter type: %r' % (args, expr),
                        ctx=ctx, param=self)
                return value
        elif self.q2_prefix == 'i':
            value = super().type_cast_value(ctx, value)
            if value is not None:
                return value[1]
            return value

        # We have an output here
        return super().type_cast_value(ctx, value)

    def _split_and_validate_input_keys(self, value):
        """ This function ensures that if a user passed in a de-facto
            collection they did so properly.
        """
        keys = [t[0] for t in value]
        values = [t[1] for t in value]

        if any(key is not None and not key.isidentifier() for key in keys):
            raise ValueError('All keys must be valid Python identifiers.'
                             ' Python identifier rules may be found here'
                             ' https://www.askpython.com/python/'
                             'python-identifiers-rules-best-practices')

        # If we had no keys, we are fine
        if all(key is None for key in keys):
            return None, values

        has_nones = any(key is None for key in keys)
        has_keys = any(key is not None for key in keys)

        # We cannot have keys for something that isn't a dict
        if self.q2_multiple is not dict and has_keys:
            raise ValueError('Keyed values may only be supplied for '
                             'Collection inputs.')
        # We cannot have a mixture of keyed and unkeyed values
        elif self.q2_multiple is dict and has_keys and has_nones:
            raise ValueError('Keyed values cannot be mixed with unkeyed '
                             'values.')

        return keys, values

    def _check_length(self, value, ctx):
        import collections

        if isinstance(value, tuple) and len(value) == 1 and \
                isinstance(value[0], dict):
            value = list(value[0].values())

        counter = collections.Counter(value)

        dups = ', '.join(map(repr, (v for v, n in counter.items() if n > 1)))
        args = ', '.join(map(repr, value))
        if dups:
            raise click.BadParameter(
                'received <%s> as an argument, which contains duplicates'
                ' of the following: <%s>' % (args, dups), ctx=ctx, param=self)



# ----------------------------------------------------------------------------
# Copyright (c) 2016-2023, QIIME 2 development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

import os
import subprocess
import tempfile

from q2cli.core.usage import CLIUsage
from q2cli.util import get_plugin_manager

import pytest


def _labeler(val):
    if hasattr(val, 'id'):
        return val.id
    return val


def get_tests():
    tests = []
    pm = get_plugin_manager()
    try:
        plugin = pm.plugins['mystery-stew']
    except KeyError:
        return tests
    for action in plugin.actions.values():
        for name in action.examples:
            tests.append((action, name))
    return tests


@pytest.mark.parametrize('action,example', get_tests(), ids=_labeler)
def test_mystery_stew(action, example):
    example_f = action.examples[example]
    use = CLIUsage(enable_assertions=True)
    example_f(use)
    rendered = '\n'.join(use.recorder)
    with tempfile.TemporaryDirectory() as tmpdir:
        for ref, data in use.get_example_data():
            data.save(os.path.join(tmpdir, ref))
        subprocess.run([rendered],
                       shell=True,
                       check=True,
                       cwd=tmpdir,
                       env={**os.environ})



# ----------------------------------------------------------------------------
# Copyright (c) 2016-2023, QIIME 2 development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

import click

import q2cli.builtin.dev
import q2cli.builtin.info
import q2cli.builtin.tools

from q2cli.click.command import BaseCommandMixin
from q2cli.core.config import CONFIG


class RootCommand(BaseCommandMixin, click.MultiCommand):
    """This class defers to either the PluginCommand or the builtin cmds"""
    _builtin_commands = {
        'info': q2cli.builtin.info.info,
        'tools': q2cli.builtin.tools.tools,
        'dev': q2cli.builtin.dev.dev
    }

    def __init__(self, *args, **kwargs):
        import re
        import sys

        unicodes = ["\u2018", "\u2019", "\u201C", "\u201D", "\u2014", "\u2013"]
        category_regex = re.compile(r'--m-(\S+)-category')

        invalid_chars = []
        categories = []
        for command in sys.argv:
            if any(x in command for x in unicodes):
                invalid_chars.append(command)

            match = category_regex.fullmatch(command)
            if match is not None:
                param_name, = match.groups()
                # Maps old-style option name to new name.
                categories.append((command, '--m-%s-column' % param_name))

        if invalid_chars or categories:
            if invalid_chars:
                msg = ("Error: Detected invalid character in: %s\nVerify the "
                       "correct quotes or dashes (ASCII) are being used."
                       % ', '.join(invalid_chars))
                click.echo(CONFIG.cfg_style('error', msg), err=True)
            if categories:
                old_to_new_names = '\n'.join(
                    'Instead of %s, trying using %s' % (old, new)
                    for old, new in categories)
                msg = ("Error: The following options no longer exist because "
                       "metadata *categories* are now called metadata "
                       "*columns* in QIIME 2.\n\n%s" % old_to_new_names)
                click.echo(CONFIG.cfg_style('error', msg), err=True)
            sys.exit(-1)

        super().__init__(*args, **kwargs)

        # Plugin state for current deployment that will be loaded from cache.
        # Used to construct the dynamic CLI.
        self._plugins = None

    @property
    def _plugin_lookup(self):
        import q2cli.util

        # See note in `q2cli.completion.write_bash_completion_script` for why
        # `self._plugins` will not always be obtained from
        # `q2cli.cache.CACHE.plugins`.
        if self._plugins is None:
            import q2cli.core.cache
            self._plugins = q2cli.core.cache.CACHE.plugins

        name_map = {}
        for name, plugin in self._plugins.items():
            if plugin['actions']:
                name_map[q2cli.util.to_cli_name(name)] = plugin
        return name_map

    def list_commands(self, ctx):
        import itertools

        # Avoid sorting builtin commands as they have a predefined order based
        # on applicability to users. For example, it isn't desirable to have
        # the `dev` command listed before `info` and `tools`.
        builtins = self._builtin_commands
        plugins = sorted(self._plugin_lookup)
        return itertools.chain(builtins, plugins)

    def get_command(self, ctx, name):
        if name in self._builtin_commands:
            return self._builtin_commands[name]

        try:
            plugin = self._plugin_lookup[name]
        except KeyError:
            from q2cli.util import get_close_matches

            possibilities = get_close_matches(name, self._plugin_lookup)
            if len(possibilities) == 1:
                hint = '  Did you mean %r?' % possibilities[0]
            elif possibilities:
                hint = '  (Possible commands: %s)' % ', '.join(possibilities)
            else:
                hint = ''

            click.echo(
                CONFIG.cfg_style('error', "Error: QIIME 2 has no "
                                 "plugin/command named %r." % name + hint),
                err=True)
            ctx.exit(2)  # Match exit code of `return None`

        return PluginCommand(plugin, name)


class PluginCommand(BaseCommandMixin, click.MultiCommand):
    """Provides ActionCommands based on available Actions"""
    def __init__(self, plugin, name, *args, **kwargs):
        import q2cli.util

        # the cli currently doesn't differentiate between methods
        # and visualizers, it treats them generically as Actions
        self._plugin = plugin
        self._action_lookup = {}
        self._hidden_actions = {}

        # Hide actions that start with _ by default
        for id, a in plugin['actions'].items():
            if id.startswith('_'):
                self._hidden_actions[q2cli.util.hidden_to_cli_name(id)] = a
            else:
                self._action_lookup[q2cli.util.to_cli_name(id)] = a

        support = 'Getting user support: %s' % plugin['user_support_text']
        website = 'Plugin website: %s' % plugin['website']
        description = 'Description: %s' % plugin['description']
        help_ = '\n\n'.join([description, website, support])

        params = [
            click.Option(('--version',), is_flag=True, expose_value=False,
                         is_eager=True, callback=self._get_version,
                         help='Show the version and exit.'),
            q2cli.util.example_data_option(self._get_plugin),
            q2cli.util.citations_option(self._get_citation_records)
        ]

        if self._hidden_actions != {}:
            params.append(
                click.Option(
                    ('--show-hidden-actions',), is_flag=True,
                    expose_value=False, is_eager=True,
                    callback=self._get_hidden_actions,
                    help="This plugin has hidden actions with names starting "
                         "with '_'. These are generally called internally by "
                         "pipelines. Passing this flag will display those "
                         "actions."))

        super().__init__(name, *args, short_help=plugin['short_description'],
                         help=help_, params=params, **kwargs)

    def _get_version(self, ctx, param, value):
        if not value or ctx.resilient_parsing:
            return

        import q2cli.util
        pm = q2cli.util.get_plugin_manager()
        for plugin in pm.plugins.values():
            if (self._plugin['name'] == plugin.name):
                pkg_name = plugin.project_name
                pkg_version = plugin.version
                break
        else:
            pkg_name = pkg_version = "[UNKNOWN]"

        click.echo(
            "QIIME 2 Plugin '%s' version %s (from package '%s' version %s)"
            % (self._plugin['name'], self._plugin['version'],
               pkg_name, pkg_version)
        )
        ctx.exit()

    def _get_citation_records(self):
        import q2cli.util
        pm = q2cli.util.get_plugin_manager()
        return pm.plugins[self._plugin['name']].citations

    def _get_hidden_actions(self, ctx, param, value):
        """Add actions that start with _ back to the lookup"""
        # Click calls this whether the flag was provided or not. If the flag
        # was not provided value is, unsurprisingly, False. We do not want to
        # execute this if the flag was not provided.
        #
        # Resilient parsing has something to do with ignoring default values
        # which for this is False. Honestly not 100% sure why we need that
        # here, but it is in the check in _get_version, and I a mimicking that
        if not value or ctx.resilient_parsing:
            return

        from click.utils import echo

        self._action_lookup.update(self._hidden_actions)

        # Handle the printing and exiting here. This feels like a pretty
        # serious misuse of click, but it probably isn't the most egregious in
        # the cli
        echo(ctx.get_help(), color=ctx.color)
        ctx.exit()

    def _get_plugin(self):
        import q2cli.util
        pm = q2cli.util.get_plugin_manager()
        return pm.plugins[self._plugin['name']]

    def list_commands(self, ctx):
        return sorted(self._action_lookup)

    def get_command(self, ctx, name):
        try:
            # Hidden actions are still valid commands
            self._action_lookup.update(self._hidden_actions)
            action = self._action_lookup[name]
        except KeyError:
            from q2cli.util import get_close_matches

            possibilities = get_close_matches(name, self._action_lookup)
            if len(possibilities) == 1:
                hint = '  Did you mean %r?' % possibilities[0]
            elif possibilities:
                hint = '  (Possible commands: %s)' % ', '.join(possibilities)
            else:
                hint = ''

            click.echo(
                CONFIG.cfg_style('error', "Error: QIIME 2 plugin %r has no "
                                 "action %r." % (self._plugin['name'], name) +
                                 hint), err=True)
            ctx.exit(2)  # Match exit code of `return None`

        return ActionCommand(name, self._plugin, action)


class ActionCommand(BaseCommandMixin, click.Command):
    """A click manifestation of a QIIME 2 API Action (Method/Visualizer)

    """
    def __init__(self, name, plugin, action):
        import q2cli.util
        import q2cli.click.type

        self.plugin = plugin
        self.action = action

        self._inputs, self._params, self._outputs = \
            self._build_generated_options()

        self._misc = [
            click.Option(['--output-dir'],
                         type=q2cli.click.type.OutDirType(),
                         help='Output unspecified results to a directory'),
            click.Option(['--verbose / --quiet'], default=None, required=False,
                         help='Display verbose output to stdout and/or stderr '
                              'during execution of this action. Or silence '
                              'output if execution is successful (silence is '
                              'golden).')
        ]

        # If this action is a pipeline it needs additional options for
        # recycling and parallelization
        action_obj = self._get_action()
        if action_obj.type == 'pipeline':
            self._misc.extend([
                click.Option(['--recycle-pool'], required=False,
                             type=str,
                             help='Use a cache pool for pipeline resumption. '
                                  'QIIME 2 will cache your results in this '
                                  'pool for reuse by future invocations. '
                                  'These pool are retained until deleted by '
                                  'the user. If not provided, QIIME 2 will '
                                  'create a pool which is automatically '
                                  'reused by invocations of the same action '
                                  'and removed if the action is successful. '
                                  'Note: these pools are local to the '
                                  'cache you are using.'),
                click.Option(['--no-recycle'], is_flag=True, required=False,
                             help='Do not recycle results from a previous '
                                  'failed pipeline run or save the results '
                                  'from this run for future recycling.'),
                click.Option(['--parallel'], is_flag=True, required=False,
                             help='Execute your action in parallel. This flag '
                                  'will use your default parallel config.'),
                click.Option(['--parallel-config'], required=False,
                             type=click.Path(exists=True, dir_okay=False),
                             help='Execute your action in parallel using a '
                                  'config at the indicated path.'),
                click.Option(['--use-cache'], required=False,
                             type=click.Path(exists=True, file_okay=False),
                             help='Specify the cache to be used for the '
                                  'intermediate work of this pipeline. If '
                                  'not provided, the default cache under '
                                  '$TMP/qiime2/<uname> will be used. '
                                  'IMPORTANT FOR HPC USERS: If you are on an '
                                  'HPC system and are using parallel '
                                  'execution it is important to set this to '
                                  'a location that is globally accessible to '
                                  'all nodes in the cluster.')])

        self._misc.extend([
            q2cli.util.example_data_option(
                self._get_plugin, self.action['id']),
            q2cli.util.citations_option(self._get_citation_records)])

        options = [*self._inputs, *self._params, *self._outputs, *self._misc]
        help_ = [action['description']]
        if self.action['deprecated']:
            help_.append(CONFIG.cfg_style(
                'warning', 'WARNING:\n\nThis command is deprecated and will '
                           'be removed in a future version of this plugin.'))
        super().__init__(name, params=options, callback=self,
                         short_help=action['name'], help='\n\n'.join(help_))

    def _build_generated_options(self):
        import q2cli.click.option

        inputs = []
        params = []
        outputs = []

        for item in self.action['signature']:
            item = item.copy()
            type = item.pop('type')

            if type == 'input':
                storage = inputs
            elif type == 'parameter':
                storage = params
            else:
                storage = outputs

            opt = q2cli.click.option.GeneratedOption(prefix=type[0], **item)
            storage.append(opt)

        return inputs, params, outputs

    def get_opt_groups(self, ctx):
        return {
            'Inputs': self._inputs,
            'Parameters': self._params,
            'Outputs': self._outputs,
            'Miscellaneous': self._misc + [self.get_help_option(ctx)]
        }

    def _get_citation_records(self):
        return self._get_action().citations

    def _get_plugin(self):
        import q2cli.util
        pm = q2cli.util.get_plugin_manager()
        return pm.plugins[self.plugin['name']]

    def _get_action(self):
        plugin = self._get_plugin()
        return plugin.actions[self.action['id']]

    def __call__(self, **kwargs):
        """Called when user hits return, **kwargs are Dict[click_names, Obj]"""
        import os
        import click

        import qiime2.util
        from q2cli.util import (output_in_cache, _get_cache_path_and_key,
                                get_default_recycle_pool)
        from qiime2.core.cache import Cache
        from qiime2.sdk import ResultCollection

        output_dir = kwargs.pop('output_dir')
        # If they gave us a cache and key combo as an output dir, we want to
        # error out, so we check if their output dir contains a : and the part
        # before it is a cache
        if output_dir:
            potential_cache = output_dir.rsplit(':', 1)[0]
            if potential_cache and os.path.exists(potential_cache) and \
                    Cache.is_cache(potential_cache):
                raise ValueError(f"The given output dir '{output_dir}' "
                                 "appears to be a cache:key combo. Cache keys "
                                 "cannot be used as output dirs.")

        # Args pertaining to pipeline resumption
        recycle_pool = kwargs.pop('recycle_pool', None)
        no_recycle = kwargs.pop('no_recycle', False)

        if recycle_pool is not None and no_recycle:
            raise ValueError('Cannot set a pool to be used for recycling and '
                             'no recycle simultaneously.')

        used_cache = kwargs.pop('use_cache', None)

        if used_cache is not None and not Cache.is_cache(used_cache):
            raise ValueError(f"The path '{used_cache}' is not a valid cache, "
                             "please supply a path to a valid pre-existing "
                             "cache.")

        parallel = kwargs.pop('parallel', False)
        parallel_config_fp = kwargs.pop('parallel_config', None)

        if parallel_config_fp is not None:
            parallel = True

        verbose = kwargs.pop('verbose')
        if verbose is None:
            verbose = False
            quiet = False
        elif verbose:
            quiet = False
        else:
            quiet = True

        cache = Cache(path=used_cache)

        arguments = {}
        init_outputs = {}
        for key, value in kwargs.items():
            prefix, *parts = key.split('_')
            key = '_'.join(parts)

            if prefix == 'o':
                if value is None:
                    value = os.path.join(output_dir, key)
                init_outputs[key] = value
            elif prefix == 'm':
                arguments[key[:-len('_file')]] = value
            # Make sure our inputs are backed by the cache we are using. This
            # is necessary for HPCs where our input .qzas may be in a location
            # that is not globally accessible to the cluster. The user should
            # be using a cache that is in a globally accessible location. We
            # need to ensure we put our artifacts in that cache.
            elif prefix == 'i' and used_cache is not None:
                value_ = value

                if isinstance(value, list):
                    value_ = [cache.process_pool.save(v) for v in value]
                elif isinstance(value, dict) or \
                        isinstance(value, ResultCollection):
                    value_ = {
                        k: cache.process_pool.save(v)
                        for k, v in value.items()}
                elif isinstance(value, set):
                    value_ = set([cache.process_pool.save(v) for v in value])
                elif value is not None:
                    value_ = cache.process_pool.save(value)

                arguments[key] = value_
            else:
                arguments[key] = value

        outputs = self._order_outputs(init_outputs)
        action = self._get_action()

        # If --no-recycle is not set, pipelines attempt to recycle their
        # outputs from a pool by default allowing recovery of failed pipelines
        # from point of failure without needing to restart the pipeline from
        # the beginning
        default_pool = get_default_recycle_pool(
            f'{action.plugin_id}_{action.id}')
        if not no_recycle and action.type == 'pipeline' and \
                recycle_pool is None:
            # We implicitly use a pool named
            # recycle_<plugin>_<action>_<sha1(plugin_action)> if no pool is
            # provided
            recycle_pool = default_pool

        if recycle_pool is not None and recycle_pool != default_pool and \
                recycle_pool not in cache.get_pools():
            msg = ("The pool '%s' does not exist on the cache at '%s'. It "
                   "will be created." % (recycle_pool, cache.path))
            click.echo(CONFIG.cfg_style('warning', msg))

        # `qiime2.util.redirected_stdio` defaults to stdout/stderr when
        # supplied `None`.
        log = None

        if not verbose:
            import tempfile
            log = tempfile.NamedTemporaryFile(prefix='qiime2-q2cli-err-',
                                              suffix='.log',
                                              delete=False, mode='w')
        if action.deprecated:
            # We don't need to worry about redirecting this, since it should a)
            # always be shown to the user and b) the framework-originated
            # FutureWarning will wind up in the log file in quiet mode.

            msg = ('Plugin warning from %s:\n\n%s is deprecated and '
                   'will be removed in a future version of this plugin.' %
                   (q2cli.util.to_cli_name(self.plugin['name']), self.name))
            click.echo(CONFIG.cfg_style('warning', msg))

        cleanup_logfile = False
        try:
            with qiime2.util.redirected_stdio(stdout=log, stderr=log):
                if parallel:
                    from qiime2.sdk.parallel_config import \
                        (get_config_from_file, ParallelConfig)

                    action = action.parallel
                    if parallel_config_fp is None:
                        parallel_config = ParallelConfig()
                    else:
                        config, mapping = \
                              get_config_from_file(parallel_config_fp)
                        parallel_config = ParallelConfig(config, mapping)

                    with parallel_config:
                        results = self._execute_action(
                            action, arguments, cache, recycle_pool)
                else:
                    results = self._execute_action(
                        action, arguments, cache, recycle_pool)
        except Exception as e:
            header = ('Plugin error from %s:'
                      % q2cli.util.to_cli_name(self.plugin['name']))
            if verbose:
                # log is not a file
                log = 'stderr'
            q2cli.util.exit_with_error(e, header=header, traceback=log)
        else:
            cleanup_logfile = True
        finally:
            # OS X will reap temporary files that haven't been touched in
            # 36 hours, double check that the log is still on the filesystem
            # before trying to delete. Otherwise this will fail and the
            # output won't be written.
            if log and cleanup_logfile and os.path.exists(log.name):
                log.close()
                os.remove(log.name)

        if output_dir is not None:
            os.makedirs(output_dir)

        for result, output in zip(results, outputs):
            if isinstance(output, tuple) and len(output) == 1:
                output = output[0]

            if output_in_cache(output) and output_dir is None:
                cache_path, key = _get_cache_path_and_key(output)
                output_cache = Cache(cache_path)

                if isinstance(result, ResultCollection):
                    output_cache.save_collection(result, key)
                    path = output
                else:
                    output_cache.save(result, key)
                    path = output
            else:
                path = result.save(output)

            if not quiet:
                if output_in_cache(output):
                    message = \
                        f"Added {result.type} to cache: {cache_path} as: {key}"
                else:
                    type = f'Collection[{list(result.values())[0].type}]' if \
                        isinstance(result, ResultCollection) else result.type
                    message = f"Saved {type} to: {path}"

                click.echo(CONFIG.cfg_style('success', message))

        # If we used a default recycle pool for a pipeline and the pipeline
        # succeeded, then we need to clean up the pool. Make sure to do this at
        # the very end so if a failure happens during writing results we still
        # have them
        if recycle_pool == default_pool:
            cache.remove(recycle_pool)

    def _execute_action(self, action, arguments, cache, recycle_pool=None):
        with cache:
            if recycle_pool is None:
                results = action(**arguments)
                results = results._result()
            else:
                pool = cache.create_pool(key=recycle_pool, reuse=True)
                with pool:
                    results = action(**arguments)

                    # If we executed in a pool using parsl we need to get
                    # our results inside of the context manager to ensure
                    # that the pool is set for the entirety of the
                    # execution
                    results = results._result()

        return results

    def _order_outputs(self, outputs):
        ordered = []
        for item in self.action['signature']:
            if item['type'] == 'output':
                ordered.append(outputs[item['name']])
        return ordered

    def format_epilog(self, ctx, formatter):
        if self.action['epilog']:
            with formatter.section(click.style('Examples', bold=True)):
                for line in self.action['epilog']:
                    formatter.write(' ' * formatter.current_indent)
                    formatter.write(line)
                    formatter.write('\n')



# ----------------------------------------------------------------------------
# Copyright (c) 2016-2023, QIIME 2 development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

# ----------------------------------------------------------------------------
# Some of the source code in this file is derived from original work:
#
# Copyright (c) 2014 by the Pallets team.
#
# To see the license for the original work, see licenses/click.LICENSE.rst
# Specific reproduction and derivation of original work is marked below.
# ----------------------------------------------------------------------------

import click.parser as parser
import click.exceptions as exceptions


class Q2Option(parser.Option):
    @property
    def takes_value(self):
        # store_maybe should take a value so that we hit the right branch
        # in OptionParser._match_long_opt
        return (super().takes_value or self.action == 'store_maybe'
                or self.action == 'append_greedy')

    def _maybe_take(self, state):
        if not state.rargs:
            return None
        # In a more perfect world, we would have access to all long opts
        # and could verify against those instead of just the prefix '--'
        if state.rargs[0].startswith('--'):
            return None
        return state.rargs.pop(0)

    # Specific technique derived from original:
    # < https://github.com/pallets/click/blob/
    #   c6042bf2607c5be22b1efef2e42a94ffd281434c/click/core.py#L867 >
    # Copyright (c) 2014 by the Pallets team.
    def process(self, value, state):
        # actions should update state.opts and state.order

        if (self.dest in state.opts
                and self.action not in ('append', 'append_const',
                                        'append_maybe', 'append_greedy',
                                        'count')):
            raise exceptions.UsageError(
                'Option %r was specified multiple times in the command.'
                % self._get_opt_name())
        elif self.action == 'store_maybe':
            assert value == ()
            value = self._maybe_take(state)
            if value is None:
                state.opts[self.dest] = self.const
            else:
                state.opts[self.dest] = value
            state.order.append(self.obj)  # can't forget this
        elif self.action == 'append_maybe':
            value = self._maybe_take(state)
            if value is None:
                state.opts.setdefault(self.dest, []).append(self.const)
            else:
                while value is not None:
                    state.opts.setdefault(self.dest, []).append(value)
                    value = self._maybe_take(state)
            state.order.append(self.obj)  # can't forget this
        elif self.action == 'append_greedy':
            assert value == ()
            value = self._maybe_take(state)
            while value is not None:
                state.opts.setdefault(self.dest, []).append(value)
                value = self._maybe_take(state)
            state.order.append(self.obj)  # can't forget this
        elif self.takes_value and value.startswith('--'):
            # Error early instead of cascading the parse error to a "missing"
            # parameter, which they ironically did provide
            raise parser.BadOptionUsage(
                self, '%s option requires an argument' % self._get_opt_name())
        else:
            super().process(value, state)

    def _get_opt_name(self):
        if hasattr(self.obj, 'secondary_opts'):
            return ' / '.join(self.obj.opts + self.obj.secondary_opts)
        if hasattr(self.obj, 'get_error_hint'):
            return self.obj.get_error_hint(None)
        return ' / '.join(self._long_opts)


class Q2Parser(parser.OptionParser):
    # Modified from original:
    # < https://github.com/pallets/click/blob/
    #   ic6042bf2607c5be22b1efef2e42a94ffd281434c/click/parser.py#L228 >
    # Copyright (c) 2014 by the Pallets team.
    def add_option(self, obj, opts, dest, action=None, nargs=1, const=None):
        """Adds a new option named `dest` to the parser.  The destination
        is not inferred (unlike with optparse) and needs to be explicitly
        provided.  Action can be any of ``store``, ``store_const``,
        ``append``, ``appnd_const`` or ``count``.
        The `obj` can be used to identify the option in the order list
        that is returned from the parser.
        """
        opts = [parser.normalize_opt(opt, self.ctx) for opt in opts]

        # BEGIN MODIFICATIONS
        if action == 'store_maybe' or action == 'append_maybe':
            # Specifically target this branch:
            # < https://github.com/pallets/click/blob/
            #   c6042bf2607c5be22b1efef2e42a94ffd281434c/click/parser.py#L341 >
            # this happens to prevents click from reading any arguments itself
            # because it will only "pop" off rargs[:0], which is nothing
            nargs = 0
            if const is None:
                raise ValueError("A 'const' must be provided when action is "
                                 "'store_maybe' or 'append_maybe'")
        elif action == 'append_greedy':
            nargs = 0

        option = Q2Option(obj=obj, opts=opts, dest=dest, action=action,
                          nargs=nargs, const=const)
        # END MODIFICATIONS
        self._opt_prefixes.update(option.prefixes)
        for opt in option._short_opts:
            self._short_opt[opt] = option
        for opt in option._long_opts:
            self._long_opt[opt] = option

    def parse_args(self, args):
        backup = args.copy()  # args will be mutated by super()
        try:
            return super().parse_args(args)
        except exceptions.UsageError:
            if '--help' in backup:
                # all is forgiven
                return {'help': True}, [], ['help']
            raise

    # Override of private member:
    # < https://github.com/pallets/click/blob/
    #   ic6042bf2607c5be22b1efef2e42a94ffd281434c/click/parser.py#L321 >
    def _match_long_opt(self, opt, explicit_value, state):
        if opt not in self._long_opt:
            from q2cli.util import get_close_matches
            # This is way better than substring matching
            possibilities = get_close_matches(opt, self._long_opt)
            raise exceptions.NoSuchOption(opt, possibilities=possibilities,
                                          ctx=self.ctx)

        return super()._match_long_opt(opt, explicit_value, state)
